"""Implementations for computations on coordinates.

This module holds all the implementations and wrappers to handle
coordinate computations in this package.

A few of the most useful classes are lifted to the main package namespace.
They should be accessed from there when used externally, but from here when
used internally.

.. autosummary::

    uwacan.Position
    uwacan.Track
    uwacan.sensor

Other coordinate wrappers
-------------------------
.. autosummary::
    :toctree: generated

    Coordinates
    Positions
    Line
    BoundingBox
    Sensor
    SensorPosition
    SensorArray
    SensorArrayPosition
    SensorArrayPositions

Unit conversions
----------------
.. autosummary::
    :toctree: generated

    nm_to_m
    m_to_nm
    mps_to_knots
    knots_to_kmph
    kmph_to_knots
    wrap_angle
    wgs84_to_utm
    utm_to_wgs84
    wgs84_to_local_transverse_mercator
    local_transverse_mercator_to_wgs84
    wgs84_to_sweref99
    sweref99_to_wgs84

Geodesic computations
---------------------
.. autosummary::
    :toctree: generated

    distance_to
    bearing_to
    shift_position
    average_angle
    angle_between

"""

import re
import numpy as np
import xarray as xr
from . import _core
from pathlib import Path
import functools


def nm_to_m(nm):
    """Convert nautical miles to meters."""
    return nm * 1852


def m_to_nm(m):
    """Convert meters to nautical miles."""
    return m / 1852


def mps_to_knots(mps):
    """Convert meters per second to knots."""
    return mps * (3600 / 1852)


def knots_to_mps(knots):
    """Convert knots to meters per second."""
    return knots * (1852 / 3600)


def knots_to_kmph(knots):
    """Convert knots to kilometers per hour."""
    return knots * 1.852


def kmph_to_knots(kmph):
    """Convert kilometers per hour to knots."""
    return kmph / 1.852


def wrap_angle(degrees):
    """Wrap an angle to (-180, 180]."""
    return 180 - np.mod(180 - degrees, 360)


_WGS84_equatorial_radius = 6_378_137.0
_WGS84_polar_radius = 6_356_752.3
_WGS84_square_compression = (_WGS84_polar_radius / _WGS84_equatorial_radius) ** 2
_mercator_scale_factor = 0.9996
_WGS84_flattening = 1 / 298.257223563
_WGS84_third_flattening = _WGS84_flattening / (2 - _WGS84_flattening)  # n on wikipedia
_WGS84_meridian_length = (
    _WGS84_equatorial_radius
    / (1 + _WGS84_third_flattening)
    * (1 + _WGS84_third_flattening**2 / 4 + _WGS84_third_flattening**4 / 64)
)  # A on wikipedia
_WGS84_UTM_alpha = [
    1 / 2 * _WGS84_third_flattening - 2 / 3 * _WGS84_third_flattening**2 + 5 / 16 * _WGS84_third_flattening**3,
    13 / 48 * _WGS84_third_flattening**2 - 3 / 5 * _WGS84_third_flattening**3,
    61 / 240 * _WGS84_third_flattening**3,
]
_WGS84_UTM_beta = [
    1 / 2 * _WGS84_third_flattening - 2 / 3 * _WGS84_third_flattening**2 + 37 / 96 * _WGS84_third_flattening**3,
    1 / 48 * _WGS84_third_flattening**2 + 1 / 15 * _WGS84_third_flattening**3,
    17 / 480 * _WGS84_third_flattening**3,
]

_WGS84_UTM_delta = [
    2 * _WGS84_third_flattening - 2 / 3 * _WGS84_third_flattening**2 - 2 * _WGS84_third_flattening**3,
    7 / 3 * _WGS84_third_flattening**2 - 8 / 5 * _WGS84_third_flattening**3,
    56 / 15 * _WGS84_third_flattening**3,
]


def _utm_zone(longitude, rounded=True):
    # Not using our `wrap_angle`` since we want [-180, 180)
    longitude = np.mod(longitude + 180, 360) - 180
    utm_zone = (longitude + 180) / 6 + 1
    if rounded:
        utm_zone = np.floor(utm_zone).astype(int)
    else:
        # The central meridian is 3° east of the indicated zone.
        utm_zone -= 3 / 6
    return utm_zone


def wgs84_to_utm(latitude, longitude, utm_zone=None):
    """Convert WGS84 latitude and longitude to UTM coordinates.

    Parameters
    ----------
    latitude : float, array_like
        The WGS84 latitude in degrees, positive on the northern hemisphere and negative on the souther hemisphere.
    longitude : float, array_like
        The WGS84 longitude in degrees, positive is east and negative is west of the prime meridian.
    utm_zone : numeric, optional
        A fixed utm zone to use. Omit this argument and the correct utm zone is chosen.

    Returns
    -------
    easting : float, array_like
        Meters easting.
    northing : float, array_like
        Meters northing.
    utm_zone : int, array_like
        The utm zone for the outputs.

    Notes
    -----
    The UTM zones are 6° wide, with zone 31 covering [0°,6°), with higher zone numbers to the east.
    The easting values are defined as meters east of 500km west of the central meridian in each zone.
    The northing values are meters north of the equator on the northern hemisphere,
    and meters north of 10,000km on the southern hemisphere.
    This implementation uses formulas from the `Wikipedia article <https://en.m.wikipedia.org/wiki/Universal_Transverse_Mercator_coordinate_system>`_.

    """
    if utm_zone is None:
        utm_zone = _utm_zone(longitude)
    central_longitude = (utm_zone - 1) * 6 - 180 + 3

    # Convert latitude and longitude to radians
    lat_rad = np.radians(latitude)
    lon_rad = np.radians(longitude)
    central_lon_rad = np.radians(central_longitude)

    # Calculate intermediate values
    t = np.sinh(
        np.arctanh(np.sin(lat_rad))
        - (2 * np.sqrt(_WGS84_third_flattening) / (1 + _WGS84_third_flattening))
        * np.arctanh((2 * np.sqrt(_WGS84_third_flattening) / (1 + _WGS84_third_flattening)) * np.sin(lat_rad))
    )
    xi_prime = np.arctan(t / np.cos(lon_rad - central_lon_rad))
    eta_prime = np.arctanh(np.sin(lon_rad - central_lon_rad) / np.sqrt(1 + t**2))

    easting = 500_000 + _mercator_scale_factor * _WGS84_meridian_length * (
        eta_prime
        + sum(
            alpha * np.cos(2 * j * xi_prime) * np.sinh(2 * j * eta_prime)
            for j, alpha in enumerate(_WGS84_UTM_alpha, start=1)
        )
    )
    northing = (latitude < 0) * 10_000_000 + _mercator_scale_factor * _WGS84_meridian_length * (
        xi_prime
        + sum(
            alpha * np.sin(2 * j * xi_prime) * np.cosh(2 * j * eta_prime)
            for j, alpha in enumerate(_WGS84_UTM_alpha, start=1)
        )
    )

    return easting, northing, utm_zone


def utm_to_wgs84(easting, northing, utm_zone, is_south=False):
    """Convert UTM coordinates to WGS84 coordinates.

    Parameters
    ----------
    easting : float, array_like
        Meters easting.
    northing : float, array_like
        Meters northing.
    utm_zone : int, array_like
        The utm zone for the coordinate.
    is_south : bool, array_like, default=False
        If the coordinate(s) are on the souther hemisphere or not.

    Returns
    -------
    latitude : float, array_like
        The WGS84 latitude in degrees, positive on the northern hemisphere and negative on the souther hemisphere.
    longitude : float, array_like
        The WGS84 longitude in degrees, positive is east and negative is west of the prime meridian.

    Notes
    -----
    The UTM zones are 6° wide, with zone 31 covering [0°,6°), with higher zone numbers to the east.
    The easting values are defined as meters east of 500km west of the central meridian in each zone.
    The northing values are meters north of the equator on the northern hemisphere,
    and meters north of 10,000km on the southern hemisphere.
    This implementation uses formulas from the `Wikipedia article <https://en.m.wikipedia.org/wiki/Universal_Transverse_Mercator_coordinate_system>`_.

    """
    xi = (northing - is_south * 10_000_000) / (_mercator_scale_factor * _WGS84_meridian_length)
    eta = (easting - 500_000) / (_mercator_scale_factor * _WGS84_meridian_length)

    xi_prime = xi - sum(
        beta * np.sin(2 * j * xi) * np.cosh(2 * j * eta) for j, beta in enumerate(_WGS84_UTM_beta, start=1)
    )
    eta_prime = eta - sum(
        beta * np.cos(2 * j * xi) * np.sinh(2 * j * eta) for j, beta in enumerate(_WGS84_UTM_beta, start=1)
    )
    chi = np.arcsin(np.sin(xi_prime) / np.cosh(eta_prime))
    latitude = np.degrees(chi + sum(delta * np.sin(2 * j * chi) for j, delta in enumerate(_WGS84_UTM_delta, start=1)))
    central_longitude = (utm_zone - 1) * 6 - 180 + 3
    longitude = central_longitude + np.degrees(np.arctan(np.sinh(eta_prime) / np.cos(xi_prime)))

    return latitude, longitude


def wgs84_to_sweref99(latitude, longitude):
    """Convert WGS84 coordinates to Sweref99.

    The Sweref99 coordinate system is the same as UTM coordinates,
    but always using UTM zone 33, as described by
    `Lantmäteriet <https://www.lantmateriet.se/sv/geodata/gps-geodesi-och-swepos/Om-geodesi/Kartprojektioner/UTM/>`_.

    Parameters
    ----------
    latitude : float, array_like
        The WGS84 latitude in degrees, positive on the northern hemisphere and negative on the souther hemisphere.
    longitude : float, array_like
        The WGS84 longitude in degrees, positive is east and negative is west of the prime meridian.

    Returns
    -------
    easting : float, array_like
        Meters easting.
    northing : float, array_like
        Meters northing.
    """
    return wgs84_to_utm(latitude=latitude, longitude=longitude, utm_zone=33)[:2]


def sweref99_to_wgs84(easting, northing):
    """Convert Sweref99 coordinates to WGS84.

    The Sweref99 coordinate system is the same as UTM coordinates,
    but always using UTM zone 33, as described by
    `Lantmäteriet <https://www.lantmateriet.se/sv/geodata/gps-geodesi-och-swepos/Om-geodesi/Kartprojektioner/UTM/>`_.

    Parameters
    ----------
    easting : float, array_like
        Meters easting.
    northing : float, array_like
        Meters northing.

    Returns
    -------
    latitude : float, array_like
        The WGS84 latitude in degrees, positive on the northern hemisphere and negative on the souther hemisphere.
    longitude : float, array_like
        The WGS84 longitude in degrees, positive is east and negative is west of the prime meridian.
    """
    return utm_to_wgs84(easting=easting, northing=northing, utm_zone=33, is_south=False)


def wgs84_to_local_transverse_mercator(latitude, longitude, reference_latitude, reference_longitude):
    """Convert WGS84 coordinates into a local transverse mercator projection.

    A local transverse mercator projection is a coordinate system measuring meters east and north
    of a fixed point.

    Parameters
    ----------
    latitude : float, array_like
        The WGS84 latitude in degrees, positive on the northern hemisphere and negative on the souther hemisphere.
    longitude : float, array_like
        The WGS84 longitude in degrees, positive is east and negative is west of the prime meridian.
    reference_latitude : float
        The WGS84 latitude of the fixed reference point.
    reference_longitude : float
        The WGS84 longitude of the fixed reference point.

    Returns
    -------
    easting : float, array_like
        Meters easting.
    northing : float, array_like
        Meters northing.
    """
    utm_zone = _utm_zone(reference_longitude, rounded=False)
    east_0, north_0, _ = wgs84_to_utm(reference_latitude, reference_longitude, utm_zone=utm_zone)
    east, north, _ = wgs84_to_utm(latitude, longitude, utm_zone=utm_zone)
    return east - east_0, north - north_0


def local_transverse_mercator_to_wgs84(easting, northing, reference_latitude, reference_longitude):
    """Convert local transverse mercator coordinates into WGS84.

    A local transverse mercator projection is a coordinate system measuring meters east and north
    of a fixed point.

    Parameters
    ----------
    easting : float, array_like
        Meters easting.
    northing : float, array_like
        Meters northing.
    reference_latitude : float
        The WGS84 latitude of the fixed reference point.
    reference_longitude : float
        The WGS84 longitude of the fixed reference point.

    Returns
    -------
    latitude : float, array_like
        The WGS84 latitude in degrees, positive on the northern hemisphere and negative on the souther hemisphere.
    longitude : float, array_like
        The WGS84 longitude in degrees, positive is east and negative is west of the prime meridian.
    """
    utm_zone = _utm_zone(reference_longitude, rounded=False)
    east_0, north_0, _ = wgs84_to_utm(reference_latitude, reference_longitude, utm_zone=utm_zone)
    return utm_to_wgs84(
        easting=easting + east_0, northing=northing + north_0, utm_zone=utm_zone, is_south=reference_latitude < 0
    )


def _geodetic_to_geocentric(lat):
    r"""Compute the geocentric latitude from geodetic, in radians.

    The geocentric latitude is the latitude as seen from the center
    of the earth. The geodetic latitude is the angle formed with the
    equatorial plane when drawing the normal at a surface on the earth.

    The conversion for the geocentric latitude :math:`\hat φ` is

    .. math::
        \hat φ = \arctan(\tan(φ) b^2/a^2)

    with the geodetic latitude :math:`φ`, and the equatorial and polar
    earth radii :math:`a, b` respectively.
    """
    return np.arctan(np.tan(lat) * _WGS84_square_compression)


def _geocentric_to_geodetic(lat):
    r"""Compute the geodetic latitude from geocentric, in radians.

    The geodetic latitude is the angle formed with the
    equatorial plane when drawing the normal at a surface on the earth.
    The geocentric latitude is the latitude as seen from the center
    of the earth.

    The conversion for the geocentric latitude :math:`φ` is

    .. math::
         φ = \arctan(\tan(\hat φ) a^2/b^2)

    with the geocentric latitude :math:`\hat φ`, and the equatorial and polar
    earth radii :math:`a, b` respectively.
    """
    return np.arctan(np.tan(lat) / _WGS84_square_compression)


def _local_earth_radius(lat):
    r"""Compute the earth radius at a given latitude, in radians.

    The formula is

    .. math::
        R( φ) = \sqrt{\frac{(a^2\cos φ)^2+(b^2\sin φ)^2}{(a\cos φ)^2+(b\sin φ)^2}}

    with the geodetic latitude :math:`φ`, and the equatorial and polar earth radii
    :math:`a, b` respectively, see https://en.wikipedia.org/wiki/Earth_radius#Location-dependent_radii.
    """
    return (
        ((_WGS84_equatorial_radius**2 * np.cos(lat)) ** 2 + (_WGS84_polar_radius**2 * np.sin(lat)) ** 2)
        / ((_WGS84_equatorial_radius * np.cos(lat)) ** 2 + (_WGS84_polar_radius * np.sin(lat)) ** 2)
    ) ** 0.5


def _haversine(theta):
    r"""Compute the haversine of an angle, in radians.

    This is the same as

    .. math:: \sin^2(θ/2)
    """
    return np.sin(theta / 2) ** 2


def distance_to(lat_1, lon_1, lat_2, lon_2):
    r"""Calculate the distance between two coordinates.

    Conventions here are λ as the longitude and φ as the latitude.
    This implementation uses the Haversine formula:

    .. math::
        c &= H(Δφ) + (1 - H(Δφ) - H(φ_1 + φ_2)) ⋅ H(Δλ)\\
        d &= 2 R(φ) ⋅ \arcsin(\sqrt{c})\\
        H(θ) &= \sin^2(θ/2)

    implemented internally with conversions to geocentric coordinates
    and a latitude-dependent earth radius.

    Parameters
    ----------
    lat_1 : float
        Latitude of the first point in degrees.
    lon_1 : float
        Longitude of the first point in degrees.
    lat_2 : float
        Latitude of the second point in degrees.
    lon_2 : float
        Longitude of the second point in degrees.

    Returns
    -------
    float
        Distance between the two points in meters.

    """
    lat_1 = np.radians(lat_1)
    lon_1 = np.radians(lon_1)
    lat_2 = np.radians(lat_2)
    lon_2 = np.radians(lon_2)
    r = _local_earth_radius((lat_1 + lat_2) / 2)
    lat_1 = _geodetic_to_geocentric(lat_1)
    lat_2 = _geodetic_to_geocentric(lat_2)
    central_angle = _haversine(lat_2 - lat_1) + (
        1 - _haversine(lat_1 - lat_2) - _haversine(lat_1 + lat_2)
    ) * _haversine(lon_2 - lon_1)
    d = 2 * r * np.arcsin(central_angle**0.5)
    return d


def bearing_to(lat_1, lon_1, lat_2, lon_2):
    r"""Calculate the heading from one coordinate to another.

    Conventions here are λ as the longitude and φ as the latitude.
    The implementation is based on spherical trigonometry, with
    conversions to geocentric coordinates.
    This can be written as

    .. math::
        Δx &= \cos(φ_1) \sin(φ_2) - \sin(φ_1)\cos(φ_2)\cos(φ_2 - φ_1) \\
        Δy &= \sin(λ_2 - λ_1)\cos(φ_2) \\
        θ &= \arctan(Δy/Δx)

    Parameters
    ----------
    lat_1 : float
        Latitude of the first point in degrees.
    lon_1 : float
        Longitude of the first point in degrees.
    lat_2 : float
        Latitude of the second point in degrees.
    lon_2 : float
        Longitude of the second point in degrees.

    Returns
    -------
    float
        Bearing from the first point to the second point in degrees, wrapped to (-180, 180].
    """
    lat_1 = np.radians(lat_1)
    lon_1 = np.radians(lon_1)
    lat_2 = np.radians(lat_2)
    lon_2 = np.radians(lon_2)
    lat_1 = _geodetic_to_geocentric(lat_1)
    lat_2 = _geodetic_to_geocentric(lat_2)

    dy = np.sin(lon_2 - lon_1) * np.cos(lat_2)
    dx = np.cos(lat_1) * np.sin(lat_2) - np.sin(lat_1) * np.cos(lat_2) * np.cos(lat_2 - lat_1)

    bearing = np.arctan2(dy, dx)
    return wrap_angle(np.degrees(bearing))


def shift_position(lat, lon, distance, bearing):
    r"""Shifts a position given by latitude and longitude by a certain distance and bearing.

    The implementation is based on spherical trigonometry, with internal
    conversions to geocentric coordinates, and using the local radius of the earth.
    This is expressed as

    .. math::
        φ_2 &= \arcsin(\sin(φ_1) ⋅ \cos(δ) + \cos(φ_1) ⋅ \sin(δ) ⋅ \cos(θ)) \\
        λ_2 &= λ_1 + \arctan(\frac{\sin(θ) ⋅ \sin(δ) ⋅ \cos(φ_1)}{\cos(δ) - \sin(φ_1) ⋅ \sin(φ_2)})

    where: φ is latitude, λ is longitude, θ is the bearing (clockwise from north),
    δ is the angular distance d/R; d being the distance traveled, R the earth's radius.

    Parameters
    ----------
    lat : float
        Latitude of the initial position in degrees.
    lon : float
        Longitude of the initial position in degrees.
    distance : float
        Distance to move from the initial position in meters.
    bearing : float
        Direction to move from the initial position in degrees.

    Returns
    -------
    new_lat : float
        Latitude of the new position in degrees.
    new_lon : float
        Longitude of the new position in degrees.
    """
    lat = np.radians(lat)
    lon = np.radians(lon)
    bearing = np.radians(bearing)
    r = _local_earth_radius(lat)
    lat = _geodetic_to_geocentric(lat)
    dist = distance / r  # angular distance
    new_lat = np.arcsin(np.sin(lat) * np.cos(dist) + np.cos(lat) * np.sin(dist) * np.cos(bearing))
    new_lon = lon + np.arctan2(
        np.sin(bearing) * np.sin(dist) * np.cos(lat), np.cos(dist) - np.sin(lat) * np.sin(new_lat)
    )
    new_lat = _geocentric_to_geodetic(new_lat)
    return np.degrees(new_lat), np.degrees(new_lon)


def average_angle(angle, resolution=None):
    """Calculate the average angle and optionally round it to a specified resolution.

    Parameters
    ----------
    angle : array_like
        Array of angles in degrees to be averaged.
    resolution : int, str, optional
        Specifies the resolution for rounding the angle. It can be an integer specifying the number
        of divisions (e.g., 4, 8, 16) or a string ('4', '8', '16', 'four', 'eight', 'sixteen').

    Returns
    -------
    float or str
        If resolution is None, returns the average angle in degrees.
        If resolution is an integer, returns the average angle rounded to this fraction of a turn.
        If resolution is a string, returns the closest named direction (e.g., 'North', 'Southwest').

    Raises
    ------
    ValueError
        If an unknown resolution specifier is provided.

    Notes
    -----
    The function converts the input angles to complex numbers, computes their mean, and then converts back to an angle.
    If a string resolution is specified, the function maps the average angle to the nearest named direction.

    Examples
    --------
    >>> average_angle([350, 10, 40, 40])
    20.15962133607971
    >>> average_angle([350, 10, 30], resolution=10)
    36.0
    >>> average_angle([350, 10, 30], resolution='four')
    'North'
    >>> average_angle([350, 10, 20], resolution='sixteen')
    'North-northeast'
    """
    complex_angle = np.exp(1j * np.radians(angle))
    angle = wrap_angle(np.degrees(np.angle(complex_angle.mean())))
    if resolution is None:
        return angle

    if not isinstance(resolution, str):
        return wrap_angle(np.round(angle / 360 * resolution) * 360 / resolution)

    resolution = resolution.lower()
    if "4" in resolution or "four" in resolution:
        resolution = 4
    elif "8" in resolution or "eight" in resolution:
        resolution = 8
    elif "16" in resolution or "sixteen" in resolution:
        resolution = 16
    else:
        raise ValueError(f"Unknown resolution specifier '{resolution}'")

    names = [
        (-180.0, "south"),
        (-90.0, "west"),
        (0.0, "north"),
        (90.0, "east"),
        (180.0, "south"),
    ]

    if resolution >= 8:
        names.extend(
            [
                (-135.0, "southwest"),
                (-45.0, "northwest"),
                (45.0, "northeast"),
                (135.0, "southeast"),
            ]
        )
    if resolution >= 16:
        names.extend(
            [
                (-157.5, "south-southwest"),
                (-112.5, "west-southwest"),
                (-67.5, "west-northwest"),
                (-22.5, "north-northwest"),
                (22.5, "north-northeast"),
                (67.5, "east-northeast"),
                (112.5, "east-southeast"),
                (157.5, "south-southeast"),
            ]
        )
    name = min([(abs(deg - angle), name) for deg, name in names], key=lambda x: x[0])[1]
    return name.capitalize()


def angle_between(lat, lon, lat_1, lon_1, lat_2, lon_2):
    """Calculate the angle between two coordinates, as seen from a center vertex.

    The angle is counted positive if the second point is clockwise of the first point,
    as seen from the center vertex.

    Parameters
    ----------
    lat : float
        Latitude of the center vertex in degrees.
    lon : float
        Longitude of the center vertex in degrees.
    lat_1 : float
        Latitude of the first point in degrees.
    lon_1 : float
        Longitude of the first point in degrees.
    lat_2 : float
        Latitude of the second point in degrees.
    lon_2 : float
        Longitude of the second point in degrees.

    Returns
    -------
    float
        The angle between the two points as seen from the center vertex, in degrees. The angle is normalized to the range (-180, 180].
    """
    bearing_1 = bearing_to(lat, lon, lat_1, lon_1)
    bearing_2 = bearing_to(lat, lon, lat_2, lon_2)
    return wrap_angle(bearing_2 - bearing_1)


class Coordinates(_core.DatasetWrap):
    """Container for latitude and longitude.

    This class has a number of useful methods to perform computations
    on coordinates.

    Parameters
    ----------
    coordinates : `xarray.Dataset`
        Dataset with at least "latitude" and "longitude"
    latitude : float
        A latitude value to store.
    longitude : float
        A longitude value to store.
    """

    def __init__(self, coordinates=None, /, latitude=None, longitude=None):
        if coordinates is None:
            coordinates = xr.Dataset(data_vars={"latitude": latitude, "longitude": longitude})
        if isinstance(coordinates, Coordinates):
            coordinates = coordinates._data.copy()
        super().__init__(coordinates)

    @property
    def coordinates(self):
        """The latitude and longitude for this coordinate, as `~xarray.Dataset`."""
        return self._data[["latitude", "longitude"]]

    @property
    def latitude(self):
        """The latitude for coordinate, as `~xarray.DataArray`."""
        return self._data["latitude"]

    @property
    def longitude(self):
        """The longitude for coordinate, as `~xarray.DataArray`."""
        return self._data["longitude"]

    def distance_to(self, other):
        """Calculate the distance to another coordinate.

        See `uwacan.positional.distance_to` for details on the implementation.

        Parameters
        ----------
        other : `Coordinates` or convertible
            The coordinate to calculate the distance to.
            If this is an object with ``latitude`` and ``longitude``
            attributes, they will be used. Otherwise, the
            `Position` parser will be called.

        Returns
        -------
        distance : `~xarray.DataArray`
            The distance to the other coordinate.
        """
        other = self._ensure_latlon(other)
        return distance_to(self.latitude, self.longitude, other.latitude, other.longitude)

    def bearing_to(self, other):
        """Calculate the bearing to another coordinate.

        See `uwacan.positional.bearing_to` for details on the implementation.

        Parameters
        ----------
        other : `Coordinates` or convertible
            The coordinate to calculate the bearing to.
            If this is an object with ``latitude`` and ``longitude``
            attributes, they will be used. Otherwise, the
            `Position` parser will be called.

        Returns
        -------
        bearing : `~xarray.DataArray`
            The bearing to the other coordinate.
        """
        other = self._ensure_latlon(other)
        return bearing_to(self.latitude, self.longitude, other.latitude, other.longitude)

    def shift_position(self, distance, bearing):
        """Shift this coordinate by a distance in a certain bearing.

        See `uwacan.positional.shift_position` for details on the implementation.

        Parameters
        ----------
        distance : array_like
            The distance to shift, in meters.
        bearing : array_like
            The bearing to shift towards, in degrees clockwise from true north.

        Returns
        -------
        shifted : cls
            An object of the same class as the one used for shifting,
            with modified latitude and longitude.
        """
        lat, lon = shift_position(self.latitude, self.longitude, distance, bearing)
        data = self.data.assign(latitude=lat, longitude=lon)
        return type(self)(data)

    @classmethod
    def _ensure_latlon(cls, data):
        if not (hasattr(data, "latitude") and hasattr(data, "longitude")):
            # If it doesn't have lat and long we need to construct an object which
            # has them. If we get lists of values we will get a `Position` with lat,lon
            # arrays here. This usually doesn't work, but we only need to access them,
            # which will work fine. The `Position` can handle many of the other
            # useful stuff, like strings and tuples
            data = Position(data)
        return data

    def local_length_scale(self):
        """How many nautical miles one longitude minute is.

        This gives the apparent length scale for the x-axis in
        mercator projections, i.e., cos(latitude).
        The scaleratio for an x-axis should be set to this value,
        if equal length x- and y-axes are desired, e.g.::

            xaxis=dict(
                title_text='Longitude',
                constrain='domain',
                scaleanchor='y',
                scaleratio=pos.local_length_scale(),
            ),
            yaxis=dict(
                title_text='Latitude',
                constrain='domain',
            ),
        """
        # We take the mean so that it works with subclasses with arrays, e.g., Line.
        return np.cos(np.radians(self.latitude.mean().item()))

    def _figure_template(self, lat=None, lon=None, extent=None, **kwargs):
        template = super()._figure_template(**kwargs)
        height = kwargs.get("height", 800)

        if lat is None:
            try:
                lat = self.bounding_box.center.latitude
            except AttributeError:
                lat = self.latitude.mean().item()
        if lon is None:
            try:
                lon = self.bounding_box.center.longitude
            except AttributeError:
                lon = self.longitude.mean().item()
        if extent is None:
            try:
                extent = self.bounding_box.extent() * 1.05
            except AttributeError:
                extent = 10_000

        lat = float(lat)
        lon = float(lon)

        # https://docs.mapbox.com/help/glossary/zoom-level/
        # The meters/pixel in mapbox is 40_075_016.686 * cos(lat) / 2^zoom / 512
        # Where R=40_075_016.686 is the equator circumference,
        # cos(lat) is the local length scale,
        # and 512 is the size of a single tile.
        # This means the zoom to fit a distance D over P pixels is log2(R/D P/512 cos(lat))
        zoom = np.log2(40_075_016.686 / float(extent) * height / 512 * np.cos(np.radians(lat)))
        template.layout.update(
            map=dict(
                style="carto-positron",
                zoom=zoom,
                center={"lat": lat, "lon": lon},
            ),
            height=height,
            margin={"r": 0, "t": 0, "l": 0, "b": 0},
            legend=dict(
                orientation="h",
                yanchor="bottom",
                y=1.02,
                xanchor="left",
                x=0,
            ),
        )
        return template

    def make_figure(self, lat=None, lon=None, extent=None, **kwargs):
        """Create a plotly figure, styled for this data.

        Parameters
        ----------
        lat : float, optional
            The center latitude for the map. Uses a reasonable default from the data it not given.
        lon : float, optional
            The center longitude for the map. Uses a reasonable default from the data it not given.
        extent : float, optional
            How much to cover with the map, in meters. Defaults to fit the data, or 10 km for single points.
        **kwargs : dict
            Keywords will be used for the figure layout. Some useful keywords are:

            - ``map_style`` to choose a `map style <https://plotly.com/python/reference/layout/#layout-map-style>`_.
                Builtin map styles: basic, carto-darkmatter, carto-darkmatter-nolabels, carto-positron,
                carto-positron-nolabels, carto-voyager, carto-voyager-nolabels, dark, light, open-street-map, outdoors,
                satellite, satellite-streets, streets, white-bg.
            - ``height`` and ``width`` sets the figure size in pixels.
            - ``title`` adds a top level title.

        """
        import plotly.graph_objects as go

        fig = go.Figure(
            layout=dict(template=self._figure_template(lat=lat, lon=lon, extent=extent, **kwargs), **kwargs)
        )
        return fig

    def plot(self, use_minutes=True, include_time=True, name=None, text=None, hover_data=None, **kwargs):
        """Create a plotly trace for the coordinates.

        Parameters
        ----------
        use_minutes : bool, default=True
            Uses degrees and decimal minutes for the latitude and longitude hover.
        include_time : bool, default=True
            Controls if a time value should be included in the hover.
            If a string is passed, it will be used to format the times using strftime syntax.
            Default formatting is ``"%Y-%m-%d %H:%M:%S"``.
        name : str, optional
            The name or label of this trace. Used for legend and hover.
        text : [str], optional
            A list of text labels to show on hover for each point.
        hover_data : dict, optional
            Mapping to add properties to the hover. The keys should match keys in the
            track data. The values are either ``True``, ``False``, or a d3-style formatting
            specification, e.g., ``":.3f"``.
            For data with time types, the formatting string is a strftime specification,
            defaulting to ``"%Y-%m-%d %H:%D:%S"``.
        **kwargs
            All other keywords are passed to `~plotly.graph_objects.Scattermap`.
            Useful keywords are:

            - ``mode`` to choose ``"lines"``, ``"markers"``, or ``"lines+markers"``
            - ``line_color`` and ``marker_color``

            Note that the ``hovertemplate``, ``customdata``, ``meta``, ``lat``, ``lon`` keywords will be overwritten.
        """
        import plotly.graph_objects as go

        customdata = []
        meta = []
        hovertemplate = ""

        lat = np.atleast_1d(self.latitude)
        lon = np.atleast_1d(self.longitude)

        if name is not None:
            hovertemplate += name + "<br>"

        if use_minutes:
            ns = np.where(lat > 0, "N", "S")
            latdeg, latmin = np.divmod(np.abs(lat), 1)
            latmin *= 60
            ew = np.where(lon > 0, "E", "W")
            londeg, lonmin = np.divmod(np.abs(lon), 1)
            lonmin *= 60
            customdata.extend([latdeg, latmin, ns, londeg, lonmin, ew])
            # The degrees are floats here, so we format with .0f.
            # Floats can properly handle nan, so it's better to keep them as floats.
            hovertemplate += "%{customdata[0]:.0f}º %{customdata[1]:.3f}' %{customdata[2]} %{customdata[3]:.0f}º %{customdata[4]:.3f}' %{customdata[5]}"
        else:
            hovertemplate += "%{lat:.6f}º %{lon:.6f}º"

        if include_time and hasattr(self, "time"):
            if include_time is True:
                include_time = "%Y-%m-%d %H:%M:%S"
            try:
                time = self.time.dt.strftime(include_time).data
            except AttributeError:
                hovertemplate += f"<br>%{{meta[{len(meta)}]}}"
                meta.append(self.time.data)
            else:
                if (time.size == 1):
                    hovertemplate += f"<br>%{{meta[{len(meta)}]}}"
                    meta.append(time.item())
                else:
                    hovertemplate += f"<br>%{{customdata[{len(customdata)}]}}"
                    customdata.append(np.atleast_1d(time))

        hover_data = hover_data or {}
        extra_fields = []
        hovertemplate += "<extra>"
        if text is not None:
            extra_fields.append("%{text}")
        for key, item in hover_data.items():
            if not item:
                continue
            if item is True:
                item = ""
            data = self[key]
            if np.issubdtype(data.dtype, np.datetime64):
                # Pre-format time data into strings.
                item = item or "%Y-%m-%d %H:%M:%S"
                data = data.dt.strftime(item).data
                item = ""
            if data.size == 1:
                extra_fields.append(f"{key}=%{{meta[{len(meta)}]{item}}}")
                meta.append(data)
            else:
                extra_fields.append(f"{key}=%{{customdata[{len(customdata)}]{item}}}")
                customdata.append(data)
        hovertemplate += "<br>".join(extra_fields)
        hovertemplate += "</extra>"

        customdata = np.stack(customdata, axis=1)
        kwargs["lat"] = lat
        kwargs["lon"] = lon
        kwargs["name"] = name
        kwargs["hovertemplate"] = hovertemplate
        kwargs["customdata"] = customdata
        kwargs["text"] = text
        kwargs["meta"] = meta

        trace = go.Scattermap(**kwargs)
        return trace


class Position(Coordinates):
    """Class for single positions.

    This class is designed for handling of single positions,
    and includes reasonable operations that can only be performed
    with single coordinates at once.

    This class supports multiple call signatures:

    - ``Position(str)``
        A single string which will be parsed. This handles the
        most common "readable" formats. Important that the string
        has the degree symbol °, minute symbol ', and second symbol ".
        The last two are optional, but minutes and seconds will only
        be parsed if the string includes the relevant symbols.
    - ``Position(pos)``
        A single argument with ``latitude`` and ``longitude`` attributes or keys.
    - ``Position(lat, lon)``
        A pair of two values will be interpreted as the latitude and longitude directly.
    - ``Position(lat_deg, lat_min, lon_deg, lon_min)``
        Four values will be interpreted as latitude and longitude with degrees and minutes.
    - ``Position(lat_deg, lat_min, lat_sec, lon_deg, lon_min, lon_sec)``
        Six values will be interpreted as latitude and longitude with degrees, minutes, and seconds.
    - ``Position(latitude=lat, longitude=lon)``
        Using keyword arguments for latitude and longitude is also supported.
    """

    @staticmethod
    def _parse_coordinates(*args, latitude=None, longitude=None):
        if latitude is not None and longitude is not None:
            return latitude, longitude

        if len(args) == 1:
            arg = args[0]
            try:
                return arg.latitude, arg.longitude
            except AttributeError:
                pass
            try:
                return arg["latitude"], arg["longitude"]
            except (KeyError, TypeError):
                pass
            if isinstance(arg, str):
                matches = re.match(
                    r"""((?P<latdeg>[+\-\d.]+)°?)?((?P<latmin>[\d.]+)')?((?P<latsec>[\d.]+)")?(?P<lathemi>[NS])?"""
                    r"""[,]?"""
                    r"""((?P<londeg>[+\-\d.]+)°?)?((?P<lonmin>[\d.]+)')?((?P<lonsec>[\d.]+)")?(?P<lonhemi>[EW])?""",
                    re.sub(r"\s", "", arg),
                ).groupdict()
                if not matches["latdeg"] or not matches["londeg"]:
                    raise ValueError(f"Cannot parse coordinate string '{arg}'")

                digits_to_parse = len(re.sub(r"\D", "", arg))
                digits_parsed = 0
                latitude = float(matches["latdeg"])
                lat_sign = 1 if latitude >= 0 else -1
                digits_parsed += len(re.sub(r"\D", "", matches["latdeg"]))
                longitude = float(matches["londeg"])
                lon_sign = 1 if longitude >= 0 else -1
                digits_parsed += len(re.sub(r"\D", "", matches["londeg"]))

                if matches["latmin"]:
                    latitude += lat_sign * float(matches["latmin"]) / 60
                    digits_parsed += len(re.sub(r"\D", "", matches["latmin"]))
                if matches["lonmin"]:
                    longitude += lon_sign * float(matches["lonmin"]) / 60
                    digits_parsed += len(re.sub(r"\D", "", matches["lonmin"]))

                if matches["latsec"]:
                    latitude += lat_sign * float(matches["latsec"]) / 3600
                    digits_parsed += len(re.sub(r"\D", "", matches["latsec"]))
                if matches["lonsec"]:
                    longitude += lon_sign * float(matches["lonsec"]) / 3600
                    digits_parsed += len(re.sub(r"\D", "", matches["lonsec"]))

                if not digits_parsed == digits_to_parse:
                    raise ValueError(
                        f"Could not parse coordinate string '{arg}', used only {digits_parsed} of {digits_to_parse} digits"
                    )

                if matches["lathemi"] == "S":
                    latitude = -abs(latitude)
                if matches["lonhemi"] == "W":
                    longitude = -abs(longitude)

                return latitude, longitude

            else:
                # We should never have just a single argument, try unpacking.
                (*args,) = arg

        if len(args) == 2:
            latitude, longitude = args
            return latitude, longitude
        elif len(args) == 4:
            (latitude_degrees, latitude_minutes, longitude_degrees, longitude_minutes) = args
            latitude = latitude_degrees + latitude_minutes / 60
            longitude = longitude_degrees + longitude_minutes / 60
            return latitude, longitude
        elif len(args) == 6:
            (
                latitude_degrees,
                latitude_minutes,
                latitude_seconds,
                longitude_degrees,
                longitude_minutes,
                longitude_seconds,
            ) = args
            latitude = latitude_degrees + latitude_minutes / 60 + latitude_seconds / 3600
            longitude = longitude_degrees + longitude_minutes / 60 + longitude_seconds / 3600
            return latitude, longitude
        else:
            raise TypeError(f"Undefined number of arguments for Position. {len(args)} was given, expects 2, 4, or 6.")

    def __init__(self, *args, latitude=None, longitude=None):
        if len(args) == 1 and isinstance(args[0], (type(self), xr.Dataset)):
            super().__init__(args[0])
        else:
            latitude, longitude = self._parse_coordinates(*args, latitude=latitude, longitude=longitude)
            super().__init__(latitude=latitude, longitude=longitude)

    def __repr__(self):
        return f"{type(self).__name__}({self.latitude.item():.4f}, {self.longitude.item():.4f})"

    def angle_between(self, first, second):
        """Calculate the angle between two positions, as seen from this position.

        The angle between two coordinates is the span that they cover,
        measured positive clockwise and negative anti-clockwise.

        Parameters
        ----------
        first : `~uwacan.positional.Coordinates` or convertable
            The first coordinate.
        second : `~uwacan.positional.Coordinates` or convertable
            The second coordinate.

        Returns
        -------
        angle : float or `xarray.DataArray`
            The angle between the two coordinates.
            This will be a float if both coordinates is a single `Position`,
            otherwise it will be a `DataArray`.
        """
        if not isinstance(first, Coordinates):
            first = Position(first)
        if not isinstance(second, Coordinates):
            second = Position(second)
        return angle_between(
            self.latitude,
            self.longitude,
            first.latitude,
            first.longitude,
            second.latitude,
            second.longitude,
        )


class BoundingBox:
    """Representation of a bounding box.

    Parameters
    ----------
    west : float
        Western edge of the bounding box.
    south : float
        Southern edge of the bounding box.
    east : float
        Eastern edge of the bounding box.
    north : float
        Western edge of the bounding box.
    """

    @classmethod
    def from_center_and_size(cls, latitude, longitude, width, height):
        """Create a bounding box from a center, width, and height."""
        east = shift_position(lat=latitude, lon=longitude, distance=width / 2, bearing=90)[1]
        west = shift_position(lat=latitude, lon=longitude, distance=width / 2, bearing=-90)[1]
        north = shift_position(lat=latitude, lon=longitude, distance=height / 2, bearing=0)[0]
        south = shift_position(lat=latitude, lon=longitude, distance=height / 2, bearing=180)[0]
        return cls(west=west, east=east, south=south, north=north)

    def __init__(self, west, south, east, north):
        self.west = west
        self.south = south
        self.east = east
        self.north = north

    def __repr__(self):
        return f"{type(self).__name__}({self.west}, {self.south}, {self.east}, {self.north})"

    @property
    def north_west(self):
        """The north-west corner of the bounding box."""
        return Position(self.north, self.west)

    @property
    def north_east(self):
        """The north-eest corner of the bounding box."""
        return Position(self.north, self.east)

    @property
    def south_west(self):
        """The south-west corner of the bounding box."""
        return Position(self.south, self.west)

    @property
    def south_east(self):
        """The south-east corner of the bounding box."""
        return Position(self.south, self.east)

    @property
    def center(self):
        """The center of the bounding box."""
        return Position(latitude=(self.north + self.south) / 2, longitude=(self.west + self.east) / 2)

    def __contains__(self, position):
        """Check if a position is within the bounding box.

        Parameters
        ----------
        position : Position
            The position to check.

        Returns
        -------
        bool
            True if the position is within the bounding box, False otherwise.
        """
        position = Position(position)
        if (self.west <= position.longitude <= self.east) and (self.south <= position.latitude <= self.north):
            return True

    def overlaps(self, other):
        """Check if another bounding box overlaps with this one.

        Two bounding boxes are considered overlapping if any of the
        corners in one of the two bounding boxes are within the other one.

        Parameters
        ----------
        other : `BoundingBox`
            Another `BoundingBox` object.

        Returns
        -------
        bool
            True if the bounding boxes overlap, False otherwise.
        """
        return (
            other.north_west in self
            or other.north_east in self
            or other.south_west in self
            or other.south_east in self
            or self.north_west in other
            or self.north_east in other
            or self.south_west in other
            or self.south_east in other
        )

    def extent(self):
        """Calculate the extent of this bounding box.

        The extent is the maximum of the height and width.

        Returns
        -------
        extent : float
            The extent, in meters.

        """
        center = self.center
        westing, northing = wgs84_to_local_transverse_mercator(
            latitude=self.north,
            longitude=self.west,
            reference_latitude=center.latitude.item(),
            reference_longitude=center.longitude.item(),
        )
        easting, southing = wgs84_to_local_transverse_mercator(
            latitude=self.south,
            longitude=self.east,
            reference_latitude=center.latitude.item(),
            reference_longitude=center.longitude.item(),
        )
        extent = max((northing - southing), (easting - westing))
        return extent


class Positions(Coordinates):
    """Container for arrays of coordinates."""

    def __repr__(self):
        return f"{type(self).__name__} with {self.latitude.size} points"

    @property
    def bounding_box(self):
        """A `BoundingBox` for these coordinates."""
        try:
            return self._bounding_box
        except AttributeError:
            pass
        west = self.longitude.min().item()
        east = self.longitude.max().item()
        north = self.latitude.max().item()
        south = self.latitude.min().item()
        self._bounding_box = BoundingBox(west=west, south=south, east=east, north=north)
        return self._bounding_box


class Line(Positions):
    """A simple line of coordinates."""

    @classmethod
    def at_position(cls, position, length, bearing, n_points=100, symmetric=False, dim="line"):
        """Create a line at a given position with a specified length and bearing.

        Parameters
        ----------
        position : `Position`
            The starting position of the line.
        length : float
            The total length of the line, in meters.
        bearing : float
            The bearing (direction) of the line in degrees.
        n_points : int, default=100
            The number of points to generate along the line.
        symmetric : bool, default=False
            If True, the line is centered on the given position.
        dim : str, default="line"
            The dimension name for the line points.

        Returns
        -------
        Line
            A new instance of the `Line` class representing the line at the specified position.
        """
        if symmetric:
            n_points += (n_points + 1) % 2
            distance = np.linspace(-length / 2, length / 2, n_points)
        else:
            distance = np.linspace(0, length, n_points)

        distance = xr.DataArray(distance, dims=dim)
        position = Position(position)
        lat, lon = shift_position(position.latitude, position.longitude, distance, bearing)
        return cls(latitude=lat, longitude=lon)


class Track(Positions):
    """A class representing a GPS track, which is a sequence of coordinates over time.

    Typically instances of this class are created by reading a data file, using the
    implemented classmethods.

    Parameters
    ----------
    data : xarray.Dataset
        The dataset containing the track data.
    calculate_course : bool, optional
        Whether to calculate the course of the track, by default False.
    calculate_speed : bool, optional
        Whether to calculate the speed of the track, by default False.
    """

    @classmethod
    def read_nmea_gps(cls, filepath, **kwargs):
        """Read a GPS track from an NMEA file.

        Parameters
        ----------
        filepath : str or Path
            The path to the NMEA file.
        **kwargs : dict, optional
            Additional keyword arguments passed to the class constructor.

        Returns
        -------
        Track
            A `Track` object containing the GPS data.

        Notes
        -----
        An NMEA file is a plain text file, with one NMEA message per line.
        These NMEA messages look something like::

            $GPRMC,102100.000,A,5741.4478,N,01152.4111,E,0.00,319.61,300823,,,A*60

        If you have a different header (GPRMC), it's likely that this read function will not work properly.
        """
        latitudes = []
        longitudes = []
        headings = []
        speeds = []
        times = []

        with open(filepath, "r") as f:
            for line in f:
                line = line.strip("\"'\n")
                (
                    header,
                    utc,
                    status,
                    lat,
                    lat_dir,
                    lon,
                    lon_dir,
                    speed,
                    heading,
                    date,
                    mag_var,
                    mag_var_dir,
                    mode_chsum,
                ) = line.split(",")
                times.append(_core.time_to_np(date + utc, fmt="%d%m%y%H%M%S.%f"))
                latitudes.append((int(lat[:2]) + float(lat[2:]) / 60) * (1 if lat_dir == "N" else -1))
                longitudes.append((int(lon[:3]) + float(lon[3:]) / 60) * (1 if lon_dir == "E" else -1))
                headings.append(float(heading))
                speeds.append(float(speed))
        track = xr.Dataset(
            data_vars=dict(
                latitude=xr.DataArray(latitudes, coords={"time": times}),
                longitude=xr.DataArray(longitudes, coords={"time": times}),
                heading=xr.DataArray(headings, coords={"time": times}),
                speed=xr.DataArray(speeds, coords={"time": times}, attrs={"unit": "knots"}),
            )
        ).drop_duplicates("time")
        return cls(track, **kwargs)

    @classmethod
    def read_blueflow(cls, filepath, renames=None, **kwargs):
        """Read a track from a BlueFlow file (Excel or CSV).

        Parameters
        ----------
        filepath : str or Path
            The path to the BlueFlow file.
        renames : dict, optional
            A dictionary for renaming columns, by default None.
        **kwargs : dict, optional
            Additional keyword arguments passed to the class constructor.

        Returns
        -------
        Track
            A `Track` object containing the BlueFlow data.

        """
        import pandas

        filepath = Path(filepath)
        if filepath.suffix == ".xlsx":
            data = pandas.read_excel(filepath.as_posix())
        elif filepath.suffix == ".csv":
            data = pandas.read_csv(filepath.as_posix())
        else:
            raise ValueError(f"Unknown fileformat for blueflow file '{filepath}'. Only xlsx and csv supported.")
        data = data.to_xarray()
        names = {}
        exp = r"([^\(\)\[\]]*) [\[\(]([^\(\)\[\]]*)[\]\)]"
        for key in list(data):
            name, unit = re.match(exp, key).groups()
            names[key] = name.strip()
            data[key].attrs["unit"] = unit
        data = data.rename(names)

        renames = {
            "Latitude": "latitude",
            "Longitude": "longitude",
            "Timestamp": "time",
            "Time": "time",
            "Tidpunkt": "time",
            "Latitud": "latitude",
            "Longitud": "longitude",
        } | (renames or {})

        renames = {key: value for key, value in renames.items() if key in data}
        data = data.rename(renames).set_coords("time").swap_dims(index="time").drop("index")
        if not np.issubdtype(data.time.dtype, np.datetime64):
            data["time"] = xr.apply_ufunc(np.datetime64, data.time, vectorize=True, keep_attrs=True)
        return cls(data, **kwargs)

    @classmethod
    def read_gpx(cls, filepath, **kwargs):
        """Read a GPS track from a GPX file.

        Parameters
        ----------
        filepath : str or Path
            The path to the GPX file.
        **kwargs : dict, optional
            Additional keyword arguments passed to the class constructor.

        Returns
        -------
        Track
            A `Track` object containing the GPX data.s
        """
        try:
            import gpxpy
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Module 'gpxpy' required to read gpx data")
        with open(filepath, "r") as f:
            contents = gpxpy.parse(f)
        latitudes = []
        longitudes = []
        times = []
        for point in contents.get_points_data():
            latitudes.append(point.point.latitude)
            longitudes.append(point.point.longitude)
            time = np.datetime64(int(point.point.time.timestamp() * 1e9), "ns")
            times.append(time)
        times = np.asarray(times)
        data = xr.Dataset(
            data_vars=dict(
                latitude=xr.DataArray(latitudes, coords={"time": times}),
                longitude=xr.DataArray(longitudes, coords={"time": times}),
            )
        )
        return cls(data, **kwargs)

    def __init__(self, data, calculate_course=False, calculate_speed=False):
        super().__init__(data)
        if calculate_course:
            self["course"] = self.calculate_course()
        if calculate_speed:
            self["speed"] = self.calculate_speed()

    @property
    def time(self):
        """The time coordinates for the track."""
        return self._data["time"]

    def calculate_course(self):
        """Calculate the course of the track."""
        coords = self.coordinates
        before = coords.shift(time=1).dropna("time")
        after = coords.shift(time=-1).dropna("time")
        interior_course = bearing_to(
            before.latitude,
            before.longitude,
            after.latitude,
            after.longitude,
        )
        first_course = bearing_to(
            coords.isel(time=0).latitude,
            coords.isel(time=0).longitude,
            coords.isel(time=1).latitude,
            coords.isel(time=1).longitude,
        ).assign_coords(time=coords.time[0])
        last_course = bearing_to(
            coords.isel(time=-2).latitude,
            coords.isel(time=-2).longitude,
            coords.isel(time=-1).latitude,
            coords.isel(time=-1).longitude,
        ).assign_coords(time=coords.time[-1])
        course = xr.concat([first_course, interior_course, last_course], dim="time")
        return course

    def calculate_speed(self):
        """Calculate the speed of the track."""
        coords = self.coordinates
        before = coords.shift(time=1).dropna("time")
        after = coords.shift(time=-1).dropna("time")

        distance_delta = distance_to(before.latitude, before.longitude, after.latitude, after.longitude)
        # We cannot reuse the previous shift here, since the time coordinate is not shifted there
        time_delta = (
            coords.time.shift(time=-1).dropna("time") - coords.time.shift(time=1).dropna("time")
        ) / np.timedelta64(1, "s")
        interior_speed = distance_delta / time_delta

        first_distance = distance_to(
            coords.isel(time=0).latitude,
            coords.isel(time=0).longitude,
            coords.isel(time=1).latitude,
            coords.isel(time=1).longitude,
        )
        first_time = (coords.time[1] - coords.time[0]) / np.timedelta64(1, "s")
        first_speed = (first_distance / first_time).assign_coords(time=coords.time[0])

        last_distance = distance_to(
            coords.isel(time=-2).latitude,
            coords.isel(time=-2).longitude,
            coords.isel(time=-1).latitude,
            coords.isel(time=-1).longitude,
        )
        last_time = (coords.time[-1] - coords.time[-2]) / np.timedelta64(1, "s")
        last_speed = (last_distance / last_time).assign_coords(time=coords.time[-1])
        speed = xr.concat([first_speed, interior_speed, last_speed], dim="time")
        speed.attrs["unit"] = "m/s"
        return speed

    def average_course(self, resolution=None):
        """Calculate the average course in the track.

        The average course is the average angle of the course,
        calculated to handle wrapping.
        See `average_angle` for more details.

        Parameters
        ----------
        resolution : str, int, optional
            If a numerical resolution is given, the output will be
            rounded to this many parts of a turn.
            If a string "four", "eight", or "sixteen", is given,
            the output is a string with the closest named bearing.
        """
        try:
            course = self["course"]
        except KeyError:
            course = self.calculate_course()
        return average_angle(course, resolution=resolution)

    def closest_point(self, other):
        """Get the point in this track closest to a position.

        Parameters
        ----------
        other : `Position`
            The point to compute distances to.

        Returns
        -------
        `Position`
            The closest position in this track to the input.
            Includes latitude, longitude, time, and distance.
        """
        distances = self.distance_to(other)
        idx = distances.argmin(...)
        return Position(self._data.isel(idx).assign(distance=distances.isel(idx)))

    def aspect_segments(
        self,
        reference,
        angles,
        segment_min_length=None,
        segment_min_angle=None,
        segment_min_duration=None,
    ):
        """Get time segments corresponding to specific aspect angles.

        Aspect angles are measured between the reference and cpa.

        Parameters
        ----------
        reference : Position
            Reference position from which to measure cpa.
        angles : array_like
            The aspect angles to find.
        segment_min_length : numeric, optional
            The minimum spatial extent of each segment, in meters.
        segment_min_angle : numeric, optional
            The minimum angular extent of each segment, in degrees.
        segment_min_duration : numeric, optional
            The minimum temporal extent of each window, in seconds.

        Returns
        -------
        segments : xarray.Dataset
            A dataset with coordinates:
                - segment : the angles specified to center the segments around
                - edge : ["start", "center", "stop"], indicating the start, stop, and center of segment
            and data variables:
                - latitude (segment, edge) : the latitudes for the segment
                - longitude (segment, edge) : the longitudes for the segment
                - time (segment, edge) : the times for the segment
                - aspect_angle (segment, edge) : the actual aspect angles for the segment
                - length (segment) : the spatial extent of each segment, in m
                - angle_span (segment) : the angular extent of each segment, in degrees
                - duration (segment) : the temporal extent of each segment, in seconds
        """
        track = self.coordinates
        cpa = self.closest_point(reference)

        try:
            iter(angles)
        except TypeError:
            single_segment = True
            angles = [angles]
        else:
            single_segment = False

        angles = np.sort(angles)
        track = track.assign(aspect_angle=reference.angle_between(cpa, track))
        if track.aspect_angle[0] > track.aspect_angle[-1]:
            # We want the angles to be negative before cpa and positive after
            track["aspect_angle"] *= -1

        angles = xr.DataArray(angles, coords={"segment": angles})
        center_indices = abs(angles - track.aspect_angle).argmin("time")
        segment_centers = track.isel(time=center_indices)

        # Run a check that we get the windows we want. A sane way might be to check that the
        # first and last windows are closer to their targets than the next window.
        if angles.size > 1:
            actual_first_angle = track.aspect_angle.sel(time=segment_centers.isel(segment=0).time)
            if abs(actual_first_angle - angles.isel(segment=0)) > abs(actual_first_angle - angles.isel(segment=1)):
                raise ValueError(
                    f"Could not find window centered at {angles.isel(segment=0):.1f}⁰, found at most {actual_first_angle:.1f}⁰."
                )
            actual_last_angle = track.aspect_angle.sel(time=segment_centers.isel(segment=-1).time)
            if abs(actual_last_angle - angles.isel(segment=-1)) > abs(actual_last_angle - angles.isel(segment=-2)):
                raise ValueError(
                    f"Could not find window centered at {angles.isel(segment=-1):.1f}⁰, found at most {actual_last_angle:.1f}⁰."
                )

        segments = []
        track_angles = track.aspect_angle.data
        track_lat = track.latitude.data
        track_lon = track.longitude.data
        track_time = track.time.data
        for angle, segment_center in segment_centers.groupby("segment", squeeze=False):
            segment_center = segment_center.squeeze()
            # Finding the start of the window
            # The inner loops here are somewhat slow, likely due to indexing into the xr.Dataset all the time
            # At the time of writing (2023-12-14), there seems to be no way to iterate over a dataset in reverse order.
            # The `groupby` method can be used to iterate forwards, which solves finding the end of the segment,
            # but calling `track.sel(time=slice(t, None, -1)).groupby('time')` still iterates in the forward order.
            center_idx = int(np.abs(track.time - segment_center.time).argmin())
            start_idx = center_idx
            if segment_min_angle:
                while abs(segment_center.aspect_angle - track_angles[start_idx]) < segment_min_angle / 2:
                    start_idx -= 1
                    if start_idx < 0:
                        raise ValueError(
                            f"Start of window at {angle}⁰ not found in track. Not sufficiently high angles from window center."
                        )
            if segment_min_duration:
                while (
                    abs(segment_center.time - track_time[start_idx]) / np.timedelta64(1, "s") < segment_min_duration / 2
                ):
                    start_idx -= 1
                    if start_idx < 0:
                        raise ValueError(
                            f"Start of window at {angle}⁰ not found in track. Not sufficient time from window center."
                        )
            if segment_min_length:
                while (
                    distance_to(
                        segment_center.latitude, segment_center.longitude, track_lat[start_idx], track_lon[start_idx]
                    )
                    < segment_min_length / 2
                ):
                    start_idx -= 1
                    if start_idx < 0:
                        raise ValueError(
                            f"Start of window at {angle}⁰ not found in track. Not sufficient distance from window center."
                        )
            # Finding the end of the window
            stop_idx = center_idx
            if segment_min_angle:
                while abs(segment_center.aspect_angle - track_angles[stop_idx]) < segment_min_angle / 2:
                    stop_idx += 1
                    if stop_idx == track.sizes["time"]:
                        raise ValueError(
                            f"End of window at {angle}⁰ not found in track. Not sufficiently high angles from window center."
                        )
            if segment_min_duration:
                while (
                    abs(segment_center.time - track_time[stop_idx]) / np.timedelta64(1, "s") < segment_min_duration / 2
                ):
                    stop_idx += 1
                    if stop_idx == track.sizes["time"]:
                        raise ValueError(
                            f"End of window at {angle}⁰ not found in track. Not sufficient time from window center."
                        )
            if segment_min_length:
                while (
                    distance_to(
                        segment_center.latitude, segment_center.longitude, track_lat[stop_idx], track_lon[stop_idx]
                    )
                    < segment_min_length / 2
                ):
                    stop_idx += 1
                    if stop_idx == track.sizes["time"]:
                        raise ValueError(
                            f"End of window at {angle}⁰ not found in track. Not sufficient distance from window center."
                        )

            # Creating the window and saving some attributes
            if start_idx == stop_idx:
                segments.append(segment_center.assign(length=0, angle_span=0, duration=0).reset_coords("time"))
            else:
                segment_start, segment_stop = track.isel(time=start_idx), track.isel(time=stop_idx)
                segments.append(
                    xr.concat([segment_start, segment_center, segment_stop], dim="time")
                    .assign_coords(edge=("time", ["start", "center", "stop"]))
                    .swap_dims(time="edge")
                    .assign(
                        length=distance_to(
                            segment_start.latitude,
                            segment_start.longitude,
                            segment_stop.latitude,
                            segment_stop.longitude,
                        ),
                        angle_span=segment_stop.aspect_angle - segment_start.aspect_angle,
                        duration=(segment_stop.time - segment_start.time) / np.timedelta64(1, "s"),
                    )
                    .reset_coords("time")
                )

        if single_segment:
            return segments[0]
        return xr.concat(segments, dim="segment")

    @property
    def time_window(self):
        """A `~uwacan.TimeWindow` describing when the data start and stops."""
        return _core.TimeWindow(start=self.time[0], stop=self.time[-1])

    def subwindow(self, time=None, /, *, start=None, stop=None, center=None, duration=None, extend=None):
        """Select a subset of the data over time.

        See `uwacan.TimeWindow.subwindow` for details on the parameters.
        """
        new_window = self.time_window.subwindow(
            time, start=start, stop=stop, center=center, duration=duration, extend=extend
        )
        if isinstance(new_window, _core.TimeWindow):
            start = _core.time_to_np(new_window.start)
            stop = _core.time_to_np(new_window.stop)
            return type(self)(self._data.sel(time=slice(start, stop)))
        else:
            time = _core.time_to_np(new_window)
            return self._data.sel(time=time, method="nearest")

    def resample(self, time, /, **kwargs):
        """Resample the Track at specific times or rate.

        Parameters
        ----------
        time : float or xr.DataArray
            If an `xr.DataArray` is provided, it represents the new time points to which the data will be resampled.
            If a float is provided, it represents the frequency in Hz at which to resample the data.
        **kwargs : dict
            Additional keyword arguments passed to the `xr.DataArray.interp` method for interpolation.

        Returns
        -------
        type(self)
            A new instance of the same type as ``self``, with the data resampled at the specified time points.
        """
        if not isinstance(time, xr.DataArray):
            n_samples = int(np.floor(self.time_window.duration * time))
            start_time = _core.time_to_np(self.time_window.start)
            offsets = np.arange(n_samples) * 1e9 / time
            time = start_time + offsets.astype("timedelta64[ns]")
        data = self._data.interp(time=time, **kwargs)
        new = type(self)(data)
        return new

    def correct_gps_offset(
        self, forwards=0, portwards=0, to_bow=0, to_stern=0, to_port=0, to_starboard=0, heading=None
    ):
        """Correct positions with respect to ship heading and particulars.

        The ``to_x`` parameters is the distances from the gps antenna to the ship sides.
        The ``forwards`` and ``portwards`` parameters are how much forward and to port the
        new positions should be, relative to the center of the ship.

        Parameters
        ----------
        forwards : numeric, default 0
            How much forwards to shift the positions, in meters
        portwards : numeric, default 0
            How much to port side to shift the positions, in meters
        to_bow : numeric, default 0
            The distance to the bow from the receiver, in meters
        to_stern : numeric, default 0
            The distance to the stern from the receiver, in meters
        to_port : numeric, default 0
            The distance to the port side from the receiver, in meters
        to_starboard : numeric, default 0
            The distance to the starboard side from the receiver, in meters
        heading : array_like, optional
            The headings of the ship. Must be compatible with the coordinates in this Track
            If None, the Track checks for a "heading", then a "course" in the contained data.

        Notes
        -----
        The positions will be shifted in the ``heading`` direction by ``forwards + (to_bow - to_stern) / 2``,
        and towards "port" ``heading - 90`` by ``portwards + (to_port - to_starboard) / 2``.
        Typical usage is to give the receiver position using the ``to_x`` arguments, and the desired
        acoustic reference location with the ``forwards`` and ``portwards`` arguments.
        Inserting correct values for all the ``to_x`` arguments will center the position on the ship middle, so that
        the ``forwards`` and ``portwards`` arguments are relative to the ship center. Alternatively, leave the ``to_x`` arguments
        as the default 0 and only give the desired ``forwards`` and ``portwards`` arguments.
        """
        forwards = forwards + (to_bow - to_stern) / 2
        portwards = portwards + (to_port - to_starboard) / 2
        if heading is None:
            try:
                heading = self.heading
            except AttributeError:
                heading = self.course
        new = self.shift_position(distance=forwards, bearing=heading)
        new = new.shift_position(distance=portwards, bearing=heading - 90)
        return new


def sensor(sensor, /, sensitivity=None, depth=None, position=None, latitude=None, longitude=None):
    """Collect sensor information.

    Typical sensor information is the position, sensitivity, and deployment depth.
    The position can be given as a string, a tuple, or separate longitude and latitudes.
    This factory function will return a `~uwacan.positional.Sensor` or
    `~uwacan.positional.SensorPosition` depending on if a position is provided.
    All arguments except ``position`` can be array-like matching an array-like ``sensor``,
    or single values.

    Parameters
    ----------
    sensor : str
        The label for the sensor. All sensors must have a label.
    sensitivity : float
        The sensor sensitivity, in dB re. V/Q,
        where Q is the desired physical unit.
    depth : float
        Sensor deployment depth.
    position : str or tuple
        A coordinate string, or a tuple with lat, lon information.
    latitude : float
        The latitude the sensor was deployed at.
    longitude : float
        The longitude the sensor was deployed at.

    Returns
    -------
    Sensor or SensorPosition
    """
    if isinstance(sensor, Sensor):
        sensor = sensor._data
    if isinstance(sensor, xr.Dataset):
        sensor = sensor[[key for key, value in sensor.notnull().items() if value.any()]]
    else:
        sensor = xr.Dataset(coords={"sensor": sensor})

    if position is not None:
        latitude, longitude = Position._parse_coordinates(position)

    if latitude is not None:
        if np.ndim(latitude):
            latitude = xr.DataArray(latitude, dims="sensor")
        sensor["latitude"] = latitude

    if longitude is not None:
        if np.ndim(longitude):
            longitude = xr.DataArray(longitude, dims="sensor")
        sensor["longitude"] = longitude

    if sensitivity is not None:
        if np.ndim(sensitivity):
            sensitivity = xr.DataArray(sensitivity, dims="sensor")
        sensor["sensitivity"] = sensitivity

    if depth is not None:
        if np.ndim(depth):
            depth = xr.DataArray(depth, dims="sensor")
        sensor["depth"] = depth

    return Sensor.from_dataset(sensor)


class Sensor(_core.DatasetWrap):
    """Container for sensor information.

    This class is typically not instantiated directly,
    but through the `~uwacan.sensor` factory function.
    """

    @classmethod
    def from_dataset(cls, data):  # noqa: D102
        if isinstance(data, _core.xrwrap):
            data = data.data

        if data.sensor.ndim == 0:
            if "latitude" in data and "longitude" in data:
                cls = SensorPosition
            else:
                cls = Sensor
        else:
            if "latitude" in data and "longitude" in data:
                if np.ptp(data["latitude"].values) == 0 and np.ptp(data["longitude"].values) == 0:
                    data["latitude"] = data["latitude"].mean()
                    data["longitude"] = data["longitude"].mean()
                    cls = SensorArrayPosition
                else:
                    cls = SensorArrayPositions
            else:
                cls = SensorArray
        return cls(data)

    def __repr__(self):
        sens = "" if "sensitivity" not in self._data else f", sensitivity={self['sensitivity']:.2f}"
        depth = "" if "depth" not in self._data else f", depth={self['depth']:.2f}"
        return f"{self.__class__.__name__}({self.label}{sens}{depth})"

    @property
    def label(self):
        """The label of this sensor."""
        return self._data["sensor"].item()

    def __and__(self, other):
        if not isinstance(other, (Sensor, xr.Dataset)):
            return NotImplemented
        return _core.concatenate([self, other], dim="sensor")

    def with_data(self, **kwargs):
        """Create a copy of this sensor with additional information.

        This method allows you to add new properties to an existing sensor.
        It's particularly useful for adding sensor-specific information
        such as gain or recording channel data that is not part of the core sensor
        class.

        Parameters
        ----------
        **kwargs : dict
            Additional data to add to the sensor. Each keyword argument
            corresponds to a new data variable with one of the following input types:

            - `~xarrayr.DataArray`: Must have a "sensor" dimension matching
              the existing sensor coordinate.
            - dict: Keys should be the sensor names, and the dictionary must
              contain all sensors in the existing data.
            - scalar: A single scalar value that will be assigned to all sensors.
            - array_like: An array with a length matching the "sensor" dimension.

        """
        data = self._data.copy()
        if "sensor" not in data.dims:
            data = data.expand_dims("sensor")
        for key, value in kwargs.items():
            if isinstance(value, xr.DataArray):
                if "sensor" not in value.dims:
                    raise ValueError("Cannot add xarray data without sensor dimension to sensors")
                data[key] = value
            elif isinstance(value, dict):
                data[key] = xr.DataArray(
                    [value[key] for key in data["sensor"].values], coords={"sensor": data["sensor"]}
                )
            elif np.size(value) == 1:
                data[key] = np.squeeze(value)
            elif np.size(value) != data["sensor"].size:
                raise ValueError(f"Cannot assign {np.size(value)} values to {data['sensor'].size} sensors")
            else:
                data[key] = xr.DataArray(value, coords={"sensor": data["sensor"]})
        return self.from_dataset(data.squeeze())


class SensorPosition(Sensor, Position):
    """Container for sensor information, including position.

    This class is typically not instantiated directly,
    but through the `~uwacan.sensor` factory function.
    """

    def __repr__(self):
        sens = "" if "sensitivity" not in self._data else f", sensitivity={self['sensitivity']:.2f}"
        depth = "" if "depth" not in self._data else f", depth={self['depth']:.2f}"
        return f"{self.__class__.__name__}({self.label}, latitude={self.latitude:.4f}, longitude={self.longitude:.4f}{sens}{depth})"


class SensorArray(Sensor):
    """Container for sensor information.

    This class is typically not instantiated directly,
    but through the `~uwacan.sensor` factory function, concatenating individual sensors,
    or using the ``&`` operator on sensors.
    """

    @property
    def sensors(self):
        """The stored `Sensor` objects, as a dict."""
        return {label: sensor(data.squeeze()) for label, data in self._data.groupby("sensor", squeeze=False)}

    def __repr__(self):
        return f"{self.__class__.__name__} with {self._data['sensor'].size} sensors"

    @property
    def label(self):
        """The labels of the sensors."""
        return tuple(self._data["sensor"].data)


class SensorArrayPositions(SensorArray, Positions):
    """Container for sensor information, including positions.

    This class is typically not instantiated directly,
    but through the `~uwacan.sensor` factory function, concatenating individual sensors,
    or using the ``&`` operator on sensors.
    """

    def __init__(self, data):
        SensorArray.__init__(self, data)


class SensorArrayPosition(SensorArray, Position):
    """Container for sensor information, with a single position for all.

    This class is typically not instantiated directly,
    but through the `~uwacan.sensor` factory function, concatenating individual sensors,
    or using the ``&`` operator on sensors.
    """

    def __init__(self, data):
        SensorArray.__init__(self, data)

    def __repr__(self):
        return f"{self.__class__.__name__} with {self._data['sensor'].size} sensors at ({self.latitude:.4f}, {self.longitude:.4f})"



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "uwacan"
copyright = "2024, Carl Andersson"
author = "Carl Andersson"
from uwacan import __version__ as release
release = release.split("+")[0]

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration
import sys
sys.path.insert(0, "ext")
extensions = [
    "numpydoc",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "default_literal_role",
    "sphinx_design",
    "sphinx_copybutton",
]
add_module_names = False
autosummary_generate = True
viewcode_line_numbers = True
numpydoc_show_class_members = False  # We run our own template instead

autosummary_context = {
    "skip_methods": [
        "__init__", # we document init in class
        # skip all inherited methods from mapping
        "clear",
        "get",
        "items",
        "keys",
        "pop",
        "popitem",
        "setdefault",
        "update",
        "values",
    ],
    "extra_methods": ["__call__"],
}

default_role = "py:obj"

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

intersphinx_mapping = {
    "numpy": ("https://numpy.org/doc/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy", None),
    "xarray": ("https://docs.xarray.dev/en/latest", None),
    "sounddevice": ("https://python-sounddevice.readthedocs.io/en/latest", None),
    "soundfile": ("https://python-soundfile.readthedocs.io/en/latest", None),
    "plotly": ("https://plotly.com/python-api-reference/", None),
    "whenever": ("https://whenever.readthedocs.io/en/latest/", None),
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]
html_css_files = [
    "custom.css",
]



# %%
class DefaultDict(dict):
    def __init__(self, default_value, **kwargs):
        super().__init__(**kwargs)
        self.default_value = default_value

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            return self.default_value

class DictMapper:
    def __init__(self, preprocessor):
        self.name = preprocessor.__name__
        self.preprocessor = preprocessor

    def __get__(self, owner, owner_class=None):
        data = getattr(owner, '_' + self.name)
        if isinstance(data, DefaultDict):
            return data.default_value
        return data
        # if not isinstance(owner.identifiers, list):
            # data, = data.values()
        return data

    def __set__(self, owner, data):
        try:
            # Multiple data
            data = {key: self.preprocessor(owner, val) for key, val in data.items()}
            # multiple_data = True
        except AttributeError as err:
            if str(err).endswith("object has no attribute 'items'"):
                data = self.preprocessor(owner, data)
                data = DefaultDict(data)
                # multiple_data = False
            else:
                raise

        # if isinstance(owner.identifiers, list):
        #     if multiple_data:
        #         data = {key: data[key] for key in owner.identifiers}
        #     else:
        #         data = {key: data for key in owner.identifiers}
        # else:
        #     if multiple_data:
        #         data = data[owner.identifiers]
        #     data = {owner.identifiers: data}
        setattr(owner, '_' + self.name, data)
        # setattr(owner, '_DictMapper__' + self.name, multiple_data)


class PlainDictInterface:
    def __init__(self, identifiers=None):
        self.identifiers = identifiers

    @property
    def identifiers(self):
        ids = self._identifiers
        if len(ids) == 0:
            return None
        if len(ids) == 1:
            return ids[0]
        return ids

    @identifiers.setter
    def identifiers(self, identifiers):
        if identifiers is None or isinstance(identifiers, str):
            identifiers = [identifiers]
        else:
            try:
                iter(identifiers)
            except TypeError as err:
                if str(err).endswith('object is not iterable'):
                    # identifiers = str(identifiers)
                    identifiers = [identifiers]
                else:
                    raise
                identifiers = list(identifiers)
        self._identifiers = identifiers

    @DictMapper
    def x(self, val):
        return val + ', preprocessed'

# %%
# This almost works! If we only send a single data in, we rig a tiny dict subclass which returns the same value for all keys.
# This is the "private" data, so we can use it to get properties for all the id keys in the internal code without hassle.
# If we send multiple data in, we store a normal dictionary that is used for both the public and private access.
# The only remaining issue is how we want to deal with keys that are strings but should be ints? Is this just a bad idea and we should leave all the keys as is?
# Alternate solution is to make a second dict subclass that we use when we have multiple data, which tests the key as a string as well?
i = PlainDictInterface(5)
i.x = {5: 'five', 3: 'three'}
print(i._x)
print(i.x)
print(i._x[5])

# %%
class Container:
    def __init__(self, data, multiple_data):
        print(f'__init__ in container, {data = }, {multiple_data = }')
        self.data = data
        self.multiple_data = multiple_data

    def __get__(self, owner, owner_class=None):
        print(f'__get__ in container')
        return self

    def __getitem__(self, key):
        print(f'__getitem__ in container')
        if self.multiple_data:
            return self.data[str(key)]
        return self.data

    def __setitem__(self, key, value):
        print(f'__setitem__ in container')
        if self.multiple_data:
            self.data[str(key)] = value


class Mapper:
    def __init__(self, preprocessor):
        self.name = preprocessor.__name__
        self.preprocessor = preprocessor

    def __get__(self, owner, owner_class=None):
        print(f'__get__ in Mapper: {self.name}')
        data = getattr(owner, '_' + self.name)
        return data.data

    def __set__(self, owner, data):
        print(f'__set__ in Mapper: {self.name}')
        try:
            # Multiple data
            data = {str(key): self.preprocessor(owner, val) for key, val in data.items()}
            multiple_data = True
        except AttributeError as err:
            if str(err).endswith("object has no attribute 'items'"):
                data = self.preprocessor(owner, data)
                multiple_data = False
            else:
                raise

        # if isinstance(owner.identifiers, list):
        #     if multiple_data:
        #         data = {key: data[key] for key in owner.identifiers}
        #     else:
        #         data = {key: data for key in owner.identifiers}
        # else:
        #     if multiple_data:
        #         data = data[owner.identifiers]
            # data = {owner.identifiers: data}


        container = Container(data, multiple_data)

        setattr(owner, '_' + self.name, container)


class Interface:
    def __init__(self, identifiers):
        self.identifiers = identifiers

    @property
    def identifiers(self):
        ids = self._identifiers
        if len(ids) == 0:
            return None
        if len(ids) == 1:
            return ids[0]
        return ids

    @identifiers.setter
    def identifiers(self, identifiers):
        if identifiers is None or isinstance(identifiers, str):
            identifiers = [identifiers]
        else:
            try:
                identifiers = [str(identifier) for identifier in identifiers]
            except TypeError as err:
                if str(err).endswith('object is not iterable'):
                    # identifiers = str(identifiers)
                    identifiers = [identifiers]
                else:
                    raise
                identifiers = [str(identifier) for identifier in identifiers]
        self._identifiers = identifiers

    @Mapper
    def x(self, x):
        return x


# %%
i = Interface([5, 3])
i.x = {5: 'five', 3: 'three'}
print('plain access:', i.x)
print('private access and key:', i._x[3])
i._x[4] = 'four'
i._x.data

# %%

class AltInterface:
    def __init__(self, identifiers, normal, mapped):
        self.identifiers = identifiers
        self.normal = normal
        self._mapped = mapped

    @property
    def identifiers(self):
        ids = self._identifiers
        if len(ids) == 0:
            return None
        if len(ids) == 1:
            return ids[0]
        return ids

    @identifiers.setter
    def identifiers(self, identifiers):
        if identifiers is None or isinstance(identifiers, str):
            identifiers = [identifiers]
        else:
            try:
                identifiers = [str(identifier) for identifier in identifiers]
            except TypeError as err:
                if str(err).endswith('object is not iterable'):
                    # identifiers = str(identifiers)
                    identifiers = [identifiers]
                else:
                    raise
                identifiers = [str(identifier) for identifier in identifiers]
        self._identifiers = identifiers

    def __getattr__(self, name):
        print('__getattr__ in AltInterface')
        name = '_' + name
        if name in self.__dict__:
            print('getting value')
            value = self.__dict__[name]
            if isinstance(value, dict):
                if len(value) == 0:
                    value, = value.values()
            return value
        else:
            print('raising')
            raise AttributeError()
# %%
i = AltInterface('id', 'normal', 'mapped')
print(i.identifiers)
print(f'getting normal value: {i.normal}')
print(f'getting mapped value: {i.mapped}')
print(f'getting private mapped value: {i._mapped}')
print('setting mapped value')
i._mapped = {5: 'five'}
print(i.mapped)
print(i._mapped)
print(i.missing)
# %%



import numpy as np
import collections.abc
import abc
import pendulum

def _sanitize_datetime_input(input):
    """Sanitize datetimes to the same internal format.

    This is not really an outwards-facing function. The main use-case is
    to make sure that we have `pendulum.DateTime` objects to work with
    internally.
    It's recommended that users use nice datetimes instead of strings,
    but sometimes a user will pass a string somewhere and then we'll try to
    parse it.
    """
    try:
        return pendulum.instance(input)
    except ValueError as err:
        if 'instance() only accepts datetime objects.' in str(err):
            pass
        else:
            raise
    try:
        return pendulum.from_timestamp(input)
    except TypeError as err:
        if 'object cannot be interpreted as an integer' in str(err):
            pass
        else:
            raise
    return pendulum.parse(input)


class TimePeriod:
    def __init__(self, start=None, stop=None, center=None, duration=None):
        if start is not None:
            start = _sanitize_datetime_input(start)
        if stop is not None:
            stop = _sanitize_datetime_input(stop)
        if center is not None:
            center = _sanitize_datetime_input(center)

        if None not in (start, stop):
            _start = start
            _stop = stop
            start = stop = None
        elif None not in (center, duration):
            _start = center - pendulum.duration(seconds=duration / 2)
            _stop = center + pendulum.duration(seconds=duration / 2)
            center = duration = None
        elif None not in (start, duration):
            _start = start
            _stop = start + pendulum.duration(seconds=duration)
            start = duration = None
        elif None not in (stop, duration):
            _stop = stop
            _start = stop - pendulum.duration(seconds=duration)
            stop = duration = None
        elif None not in (start, center):
            _start = start
            _stop = start + (center - start) / 2
            start = center = None
        elif None not in (stop, center):
            _stop = stop
            _start = stop - (stop - center) / 2
            stop = center = None
        else:
            raise TypeError('Needs two of the input arguments to determine time range.')

        if (start, stop, center, duration) != (None, None, None, None):
            raise TypeError('Cannot input more than two input arguments to a time window!')

        self._period = pendulum.period(_start, _stop)

    def subperiod(self, time=None, /, *, start=None, stop=None, center=None, duration=None):
        if time is None:
            # Period specified with keyword arguments, convert to period.
            if (start, stop, center, duration).count(None) == 3:
                # Only one argument which has to be start or stop, fill the other from self.
                if start is not None:
                    period = type(self)(start=start, stop=self.stop)
                elif stop is not None:
                    period = type(self)(start=self.start, stop=stop)
                else:
                    raise TypeError('Cannot create subperiod from arguments')
            else:
                # The same types explicit arguments as the normal constructor
                period = type(self)(start=start, stop=stop, center=center, duration=duration)
        elif isinstance(time, type(self)):
            period = time
        elif isinstance(time, pendulum.Period):
            period = type(self)(start=time.start, stop=time.stop)
        else:
            # It's not a period, so it shold be a single datetime. Parse or convert, check valitidy.
            time = _sanitize_datetime_input(time)
            if time not in self:
                raise ValueError(f"Received time outside of contained period")
            return time

        if period not in self:
            raise ValueError("Requested subperiod is outside contained time period")
        return period

    def __repr__(self):
        return f'TimePeriod(start={self.start}, stop={self.stop})'

    @property
    def start(self):
        return self._period.start

    @property
    def stop(self):
        return self._period.end

    @property
    def center(self):
        return self.start + pendulum.duration(seconds=self._period.total_seconds() / 2)

    def __contains__(self, other):
        if isinstance(other, type(self)):
            other = other._period
        if isinstance(other, pendulum.Period):
            return other.start in self._period and other.end in self._period
        return other in self._period


class TimeLeafin(abc.ABC):
    @property
    @abc.abstractmethod
    def time_period(self):
        ...

    @abc.abstractmethod
    def time_subperiod(self, time=None, /, *, start=None, stop=None, center=None, duration=None):
        ...


class TimeBranchin:
    def time_subperiod(self, time=None, /, *, start=None, stop=None, center=None, duration=None, **kwargs):
        children = {
            name: child.time_subperiod(time, start=start, stop=stop, center=center, duration=duration, **kwargs)
            for name, child in self.items()
        }
        return self.clone(children)


class Metadata(collections.UserDict):
    def __init__(self, *args, node=None, **metadata):
        self.node = node
        super().__init__(*args, **metadata)

    @property
    def _parent_metadata(self):
        if self.node._parent:
            return self.node._parent.metadata
        return {}

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            pass
        try:
            return self._parent_metadata[key]
        except KeyError:
            pass
        raise KeyError(f"Metadata '{key}' cannot be found for '{self.node.__class__.__name__}' object")

    def __contains__(self, key):
        return (super().__contains__(key)) or (key in self._parent_metadata)


class Node:
    def __init__(self, metadata=None):
        if isinstance(metadata, collections.abc.Mapping):
            metadata = Metadata(node=self, **metadata)
        elif metadata is None:
            metadata = Metadata(node=self)
        else:
            raise TypeError("Cannot use object of type '{metadata.__class__.__name__}' as metadata")
        self.metadata = metadata
        self._parent = None

    @property
    def _root(self):
        try:
            return self._parent._root
        except AttributeError:
            return self

    @property
    def _parent(self):
        return self.__parent

    @_parent.setter
    def _parent(self, parent):
        self.__parent = parent

    def apply(self, function, *args, **kwargs):
        if not isinstance(function, NodeOperation):
            function = LeafDataFunction(function)
        return function(self, *args, **kwargs)

    def reduce(self, function, dim, *args, **kwargs):
        if not isinstance(function, NodeOperation):
            function = Reduction(function)
        return function(self, dim=dim, *args, **kwargs)


class Leaf(Node):
    dims = tuple()

    @property
    def _leaf_type(self):
        return type(self)

    def _traverse(self, leaves=True, branches=True, root=True, topdown=True, return_depth=False):
        if not leaves:
            return
        yield (return_depth - 1, self) if return_depth else self

    @property
    def data(self):
        return self._data


class Branch(Node, collections.abc.MutableMapping):
    def __init__(self, dim, children=None, **kwargs):
        super().__init__(**kwargs)
        self.dim = dim
        self._children = {}
        if children is not None:
            for name, child in children.items():
                self[name] = child

    def clone(self, children):
        return type(self)(dim=self.dim, children=children, metadata=self.metadata)

    def __len__(self):
        return len(self._children)

    def __iter__(self):
        return iter(self._children)

    def __getitem__(self, key):
        return self._children[key]

    def __setitem__(self, key, value):
        if key in self:
            raise KeyError(f"Cannot overwrite data '{key}' in dimension {self.dim}")
        value._parent = self
        self._children[key] = value

    def __delitem__(self, key):
        if key not in self:
            raise KeyError(f"Non-existent key '{key}' cannot be removed from data in dimension {self.dim}")
        self._children[key]._parent = None
        del self._children[key]

    @property
    def _first(self):
        return next(iter(self.values()))

    @property
    def _leaf_type(self):
        return self._first._leaf_type

    def _traverse(self, leaves=True, branches=True, root=True, topdown=True, return_depth=False):
        if return_depth is True:
            return_depth = 1
        if root and topdown:
            yield self if not return_depth else (return_depth - 1, self)
        for child in self.values():
            yield from child._traverse(leaves=leaves, branches=branches, root=branches, topdown=topdown, return_depth=return_depth and return_depth + 1)
        if root and not topdown:
            yield self if not return_depth else (return_depth - 1, self)


class NodeOperation:
    def __init__(self, function):
        self.function = function

    def __call__(self, node, *args, **kwargs):
        if isinstance(node, Branch):
            children = {name: self(child, *args, **kwargs) for name, child in node.items()}
            return node.clone(children)

    @staticmethod
    def wrap_output(output, input_leaf):
        if isinstance(output, Leaf):
            new_leaf = output
            new_leaf.metadata = Metadata(input_leaf.metadata | new_leaf.metadata, node=new_leaf)
        else:
            new_leaf = input_leaf.clone(output)
        return new_leaf


class LeafDataFunction(NodeOperation):
    def __call__(self, leaf, *args, **kwargs):
        if out := super().__call__(leaf, *args, **kwargs):
            return out

        out = self.function(leaf.data, *args, **kwargs)
        return self.wrap_output(out, leaf)


class LeafFunction(NodeOperation):
    def __call__(self, leaf, *args, **kwargs):
        if out := super().__call__(leaf, *args, **kwargs):
            return out

        out = self.function(leaf, *args, **kwargs)
        return self.wrap_output(out, leaf)


class Reduction(NodeOperation):
    def __call__(self, root, dim, metadata_merger='keep equal', *args, **kwargs):
        # TODO: add something that allows these reductions to be used on reducing axes on leaves as well?
        if isinstance(root, Leaf):
            # TODO: if we want to use the same reduction function for reducing over "layers" in the tree but also reducing over the axes of a data leaf, we will have problems with the axes somewhere.
            return root.reduce(self.function, dim=dim, *args, **kwargs)

        if dim != root.dim:
            return super().__call__(root, dim, *args, metadata_merger=metadata_merger, **kwargs)

        if metadata_merger == 'keep equal':
            def metadata_merger(name, labels, data):
                try:
                    is_equal = np.allclose(data, data[0])
                except TypeError:
                    is_equal = all(x == data[0] for x in data)
                if is_equal:
                    return data[0]
        elif metadata_merger == 'stack':
            def metadata_merger(name, labels, data):
                return data
        elif metadata_merger == 'collect':
            def metadata_merger(name, labels, data):
                return {label: x for (label, x) in zip(labels, data)}

        new = root._first.copy()
        for new_node, *old_nodes in zip(
                new._traverse(leaves=True, branches=True, root=True, topdown=False, return_depth=False),
                *(
                    child._traverse(leaves=True, branches=True, root=True, topdown=False, return_depth=False)
                    for child in root.values()
                )
        ):
            if isinstance(new_node, Leaf):
                stacked = [item.data for item in old_nodes]
                # Sometimes error handling in python is extremely annoying...
                try:
                    data = self.function(stacked, *args, **kwargs, axis=0)
                except TypeError as err:
                    err_msg = str(err)
                    if not (
                        "got an unexpected keyword argument 'axis'" in err_msg
                        or "takes no keyword arguments" in err_msg
                        or "got multiple values for keyword argument 'axis'" in err_msg
                    ):
                        raise
                    try:
                        data = self.function(stacked, *args, **kwargs)
                    except Exception as err:
                        raise err from None
                new_node._data = data
            metadata = {}
            for key in new_node.metadata:
                stacked = [item.metadata[key] for item in old_nodes]
                labels = list(root)
                this_meta = metadata_merger(key, labels, stacked)
                if this_meta is not None:
                    metadata[key] = this_meta
            new_node.metadata = Metadata(node=new_node, **metadata)
        new.metadata = root.metadata | new.metadata  # This merge will promote metadata which survived the normal pruning, prioritizing the promoted data
        new.metadata.node = new
        return new



"""Handles positional tracks.

This manages logs over positions of measurement objects via e.g. GPS.
Some of the operations include smoothing data, calculating distances,
and reading log files.
"""

import abc
import numpy as np
import scipy.interpolate
import scipy.signal
from geographiclib.geodesic import Geodesic
# from . import timestamps
import datetime
import dateutil
import bisect
geod = Geodesic.WGS84


one_knot = 1.94384


def parse_timestamp(stamp):
    return dateutil.parser.parse(stamp)
    stamp = ''.join(c for c in stamp if c in '1234567890')
    num_chars = len(stamp)
    year = int(stamp[0:4])
    month = int(stamp[4:6]) if num_chars > 4 else 1
    day = int(stamp[6:8]) if num_chars > 6 else 1
    hour = int(stamp[8:10]) if num_chars > 8 else 0
    minute = int(stamp[10:12]) if num_chars > 10 else 0
    second = int(stamp[12:14]) if num_chars > 12 else 0
    microsecond = int(stamp[14:18].ljust(6, '0')) if num_chars > 12 else 0
    return datetime.datetime(year, month, day, hour, minute, second, microsecond)


class TimeWindow:
    def __init__(self, start=None, stop=None, center=None, duration=None):
        if isinstance(start, str):
            start = parse_timestamp(start)
        if isinstance(stop, str):
            stop = parse_timestamp(stop)
        if isinstance(center, str):
            center = parse_timestamp(center)

        if None not in (start, stop):
            self._start = start
            self._stop = stop
            start = stop = None
        elif None not in (center, duration):
            self._start = center - datetime.timedelta(seconds=duration / 2)
            self._stop = center + datetime.timedelta(seconds=duration / 2)
            center = duration = None
        elif None not in (start, duration):
            self._start = start
            self._stop = start + datetime.timedelta(seconds=duration)
            start = duration = None
        elif None not in (stop, duration):
            self._stop = stop
            self._start = stop - datetime.timedelta(seconds=duration)
            stop = duration = None
        elif None not in (start, center):
            self._start = start
            self._stop = start + (center - start) / 2
            start = center = None
        elif None not in (stop, center):
            self._stop = stop
            self._start = stop - (stop - center) / 2
            stop = center = None

        if (start, stop, center, duration) != (None, None, None, None):
            raise ValueError('Cannot input more than two input arguments to a time window!')

    def __repr__(self):
        return f'TimeWindow(start={self.start}, stop={self.stop})'

    @property
    def start(self):
        return self._start

    @property
    def stop(self):
        return self._stop

    @property
    def duration(self):
        return (self.stop - self.start).total_seconds()

    @property
    def center(self):
        return self.start + datetime.timedelta(seconds=self.duration / 2)

    def __contains__(self, other):
        if isinstance(other, TimeWindow):
            return (other.start >= self.start) and (other.stop <= self.stop)
        if isinstance(other, Position):
            return self.start <= other.timestamp <= self.stop
        if isinstance(other, datetime.datetime):
            return self.start <= other <= self.stop
        raise TypeError(f'Cannot check if {other.__class__.__name__} is within a time window')


class Position:
    def __init__(self, latitude, longitude, timestamp=None):
        self._latitude = latitude
        self._longitude = longitude
        self._timestamp = timestamp

    @classmethod
    def from_degrees_minutes_seconds(cls, *args):
        if len(args) == 2:
            latitude, longitude = args
        elif len(args) == 4:
            latitude_degrees, latitude_minutes, longitude_degrees, longitude_minutes = args
            latitude = latitude_degrees + latitude_minutes / 60
            longitude = longitude_degrees + longitude_minutes / 60
        elif len(args) == 6:
            latitude_degrees, latitude_minutes, latitude_seconds, longitude_degrees, longitude_minutes, longitude_seconds = args
            latitude = latitude_degrees + latitude_minutes / 60 + latitude_seconds / 3600
            longitude = longitude_degrees + longitude_minutes / 60 + longitude_seconds / 3600
        return cls(latitude=latitude, longitude=longitude)

    def __repr__(self):
        lat = f'latitude={self.latitude}'
        lon = f', longitude={self.longitude}'
        time = f', timestamp={self.timestamp}' if self.timestamp is not None else ''
        cls = self.__class__.__name__
        return cls + '(' + lat + lon + time + ')'

    @property
    def latitude(self):
        return self._latitude

    @property
    def longitude(self):
        return self._longitude

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def coordinates(self):
        return self.latitude, self.longitude

    def copy(self, deep=False):
        obj = type(self).__new__(type(self))
        obj._latitude = self._latitude
        obj._longitude = self._longitude
        obj._timestamp = self._timestamp
        return obj

    def distance_to(self, other):
        # TODO: Make sure that this broadcasts properly!
        # I expect that the geodesic doesn't broadcast, so you might need to loop. Look at np.nditer or np.verctorize
        try:
            iter(other)
        except TypeError as err:
            if str(err).endswith('object is not iterable'):
                other = [other]
            else:
                raise

        distances = [
            geod.Inverse(
                self.latitude,
                self.longitude,
                pos.latitude,
                pos.longitude,
                outmask=geod.DISTANCE,
            )['s12']
            for pos in other
        ]
        if len(distances) == 1:
            return distances[0]
        return np.asarray(distances)

    def angle_between(self, first_position, second_position):
        first_azimuth = geod.Inverse(
            self.latitude, self.longitude,
            first_position.latitude, first_position.longitude,
            outmask=geod.AZIMUTH
        )['azi1']
        second_azimuth = geod.Inverse(
            self.latitude, self.longitude,
            second_position.latitude, second_position.longitude,
            outmask=geod.AZIMUTH
        )['azi1']
        angular_difference = second_azimuth - first_azimuth
        return (angular_difference + 180) % 360 - 180


class Track(abc.ABC):
    """A track of positions measured over time.

    Parameters
    ----------
    latitude : array_like
        The measured latitudes
    longitude : array_like
        The measured longitudes
    time : `TimedSequence`
        Specification of the times where the position was measured.
    """

    def copy(self, deep=False):
        obj = type(self).__new__(type(self))
        return obj

    @abc.abstractmethod
    def __len__(self):
        ...

    @property
    @abc.abstractmethod
    def timestamps(self):
        """List of timestamps as `datetime` objects for each position."""
        ...

    @property
    @abc.abstractmethod
    def track_time(self):
        """Array of time in the track relative to the start of the track, in seconds."""
        ...

    @property
    @abc.abstractmethod
    def time_window(self):
        """Time window that the track covers."""
        ...

    @property
    @abc.abstractmethod
    def latitude(self):
        """Latitudes of the track, in degrees."""
        ...

    @property
    @abc.abstractmethod
    def longitude(self):
        """Longitudes of the track, in degrees."""
        ...

    @property
    def coordinates(self):
        return np.stack([self.latitude, self.longitude], axis=0)

    @property
    @abc.abstractmethod
    def speed(self):
        """Speed in the track, in meters per second"""
        ...

    @property
    @abc.abstractmethod
    def heading(self):
        """Heading in the track, in degrees"""
        ...

    @property
    def mean(self):
        """Mean position of the track."""
        lat = np.mean(self.latitude)
        lon = np.mean(self.longitude)
        return Position(latitude=lat, longitude=lon)

    @property
    def boundaries(self):
        """Boundaries of the track.

        (lat_min, lat_max, lon_min, lon_max)
        """
        min_lat = np.min(self.latitude)
        min_lon = np.min(self.longitude)
        max_lat = np.max(self.latitude)
        max_lon = np.max(self.longitude)
        return min_lat, max_lat, min_lon, max_lon

    def distance_to(self, other):
        if isinstance(other, Track):
            raise TypeError('Cannot calculate distances between two tracks')
        if isinstance(other, Position):
            return other.distance_to(self)
        try:
            lat, lon = other
        except TypeError as err:
            if str(err).startswith('cannot unpack non-iterable'):
                raise TypeError(f'Cannot calculate distance between track and {other}')
            else:
                raise
        return Position(latitude=lat, longitude=lon).distance_to(self)

    def closest_point(self, other):
        if not isinstance(other, Position):
            try:
                lat, lon = other
            except TypeError as err:
                if str(err).startswith('cannot unpack non-iterable'):
                    raise TypeError(f'Cannot calculate distance between track and {other}')
                else:
                    raise
            other = Position(latitude=lat, longitude=lon)
        distances = other.distance_to(self)
        idx = np.argmin(distances)
        distance = distances[idx]
        position = self[idx]
        position.distance = distance
        return position

    def time_range(self, start=None, stop=None, center=None, duration=None):
        """Restrict the time range.

        This gets a window to the same time signal over a specified time period.
        The time period can be specified with any combination of two of the input
        parameters. The times can be specified either as `datetime` objects,
        or as strings on following format: `YYYYMMDDhhmmssffffff`.
        The microsecond part can be optionally omitted, and any non-digit characters
        are removed. Some examples include
        - 20220525165717
        - 2022-05-25_16-57-17
        - 2022/05/25 16:57:17.123456

        Parameters
        ----------
        start : datetime or string
            The start of the time window
        stop : datetime or string
            The end of the time window
        center : datetime or string
            The center of the time window
        duration : numeric
            The total duration of the time window, in seconds.
        """
        time_window = TimeWindow(
            start=start,
            stop=stop,
            center=center,
            duration=duration,
        )
        return self[time_window]


    @abc.abstractmethod
    def __getitem__(self, key):
        """Restrict the time range.

        This gets the same signal but restricted to a time range specified
        with a TimeWindow object.

        Parameters
        ----------
        window : `TimeWindow`
            The time window to restrict to.
        """
        ...

    def aspect_windows(self, reference_point, resolution, span, window_min_length=None, window_min_angle=None):
        """Get time windows corresponding to specific aspect angles

        Parameters
        ----------
        reference_point : Position
            The position from where the angles are calculated.
        resolution : numeric
            The spacing of the windows, in degrees.
        span : numeric or (numeric, numeric)
            The angular span of the centers of the windows, in degrees.
            A single value will be used as (-span, span).
        window_min_length : numeric, optional
            The minimum length of each window, in meters.
        window_min_angle : numeric, optional
            The minimum length of each window, seen as an angle from the reference point.
            If neither of `window_min_length` or `window_min_angle` is given, the `window_min_angle`
            defaults to `resolution`.
        """
        # TODO: Include some check that the path is reasonable?
        cpa = self.closest_point(reference_point)  # If the path if to long this will crunch a shit-ton of data...
        try:
            min_angle, max_angle = min(span), max(span)
        except TypeError:
            min_angle, max_angle = -span, span
        # pre_cpa, post_cpa = , self[cpa.timestamp:]

        if (window_min_angle, window_min_length) == (None, None):
            window_min_angle = resolution

        pre_cpa_angles = np.arange(1, np.math.ceil(-min_angle // resolution) + 1) * resolution
        post_cpa_angles = np.arange(1, np.math.ceil(max_angle // resolution) + 1) * resolution

        pre_cpa_window_centers = [cpa]
        post_cpa_window_centers = [cpa]
        for angle in pre_cpa_angles:
            for point in reversed(self[:pre_cpa_window_centers[-1].timestamp]):
                if abs(reference_point.angle_between(cpa, point)) >= angle:
                    pre_cpa_window_centers.append(point)
                    break
        for angle in post_cpa_angles:
            for point in self[post_cpa_window_centers[-1].timestamp:]:
                if abs(reference_point.angle_between(cpa, point)) >= angle:
                    post_cpa_window_centers.append(point)
                    break
        # Merge the list of window centers
        # The pre cpa list is reversed to have them in the correct order
        # Both lists have the cpa in them, so it's removed from the post cpa list
        window_centers = pre_cpa_window_centers[::-1] + post_cpa_window_centers[1:]

        windows = []
        for center in window_centers:
            meets_angle_criteria = window_min_angle is None
            meets_length_criteria = window_min_length is None
            for point in reversed(self[:center.timestamp]):
                if not meets_angle_criteria:
                    # Calculate angle and check if it's fine
                    meets_angle_criteria = abs(reference_point.angle_between(center, point)) >= window_min_angle / 2
                if not meets_length_criteria:
                    # Calculate length and check if it's fine
                    meets_length_criteria = center.distance_to(point) >= window_min_length / 2
                if meets_length_criteria and meets_angle_criteria:
                    window_start = point
                    break
            meets_angle_criteria = window_min_angle is None
            meets_length_criteria = window_min_length is None
            for point in self[center.timestamp:]:
                if not meets_angle_criteria:
                    # Calculate angle and check if it's fine
                    meets_angle_criteria = abs(reference_point.angle_between(center, point)) >= window_min_angle / 2
                if not meets_length_criteria:
                    # Calculate length and check if it's fine
                    meets_length_criteria = center.distance_to(point) >= window_min_length / 2
                if meets_length_criteria and meets_angle_criteria:
                    window_stop = point
                    break
            windows.append(TimeWindow(start=window_start.timestamp, stop=window_stop.timestamp))
        return windows


    def resample(self, sampletime, order='linear'):
        """Resample a position track.

        Parameters
        ----------
        sampletime : numerical
            The desired time between samples of the track.
        order : integer or string, default "linear"
            Sets the polynomial order of the interpolation.
            See `kind` argument of `scipy.interpolate.interp1d`
        """
        return ResampledTrack(
            sampletime=sampletime,
            time=self.track_time,
            start_time=self.time_window.start,
            order=order,
            latitude=self.latitude,
            longitude=self.longitude,
        )
        # TODO: Move this implementation to a ResampledTrack, which takes position data and extra data as the input.
        # TODO: add this to the Blueflow class as well to make sure that any extra props get included. Check if you can make this nice in the api?

        # Handling uniqueness of samples.
        unique_times, time_indices, duplicate_counts = np.unique(self.time, return_inverse=True, return_counts=True)
        latitude = self.latitude
        longitude = self.longitude
        if unique_times.size < time_indices.size:
            # We have duplicate timestamps
            # bincount will count how often each value in `time_indices` occurs
            # i.e. how many times the same time exists in the time vector.
            # These values are weighted by the input values in latitude and longitude,
            # given as the second argument.
            # Finally the mean is calculated by dividing by the number of duplicates.
            latitude = np.bincount(time_indices, latitude) / duplicate_counts
            longitude = np.bincount(time_indices, longitude) / duplicate_counts

        # Handling subsampling of the domain
        # if start is not None:
        #     # Get the index which is just to the right of the start value, then go left one.
        #     # This is to include at one additional sample if the requested time is not actually in the times.
        #     start_idx = max(np.searchsorted(unique_times, start, 'right') - 1, 0)
        # else:
        #     start_idx = 0
        #     start = unique_times[0]

        # if stop is not None:
        #     # Get the index which is just to the left of the start value, then go right one.
        #     # This is to include at one additional sample if the requested time is not actually in the times.
        #     stop_idx = min(np.searchsorted(unique_times, stop, 'left') + 1, len(unique_times))
        # else:
        #     stop_idx = len(unique_times)
        #     stop = unique_times[-1]
        # unique_times = unique_times[start_idx:stop_idx]
        # latitude = latitude[start_idx:stop_idx]
        # longitude = longitude[start_idx:stop_idx]

        # Performing the interpolation
        n_samples = np.math.floor((unique_times[-1] - unique_times[0]) / sampletime) + 1
        time = np.arange(n_samples) * sampletime + unique_times[0]
        def interpolate(data):
            interpolator = scipy.interpolate.interp1d(
                unique_times, data,
                kind=order,
                bounds_error=False, fill_value=(data[0], data[-1])
            )
            return interpolator(time)
        latitude = interpolate(latitude)
        longitude = interpolate(longitude)

        resampled = {}
        for key in extra_data:
            resampled[key] = interpolate(getattr(self, key))
        # Store the new data
        obj = ResampledTrack(time=time, latitude=latitude, longitude=longitude, **resampled)
        return obj
        resampled._latitude = latitude
        resampled._longitude = longitude
        resampled.time = timestamps.time
        resampled.sampletime = sampletime
        return resampled


class TimestampedTrack(Track):
    def __init__(self, timestamps):
        self._timestamps = np.asarray(timestamps)

    @property
    def timestamps(self):
        return self._timestamps

    @property
    def time_window(self):
        return TimeWindow(start=self.timestamps[0], stop=self.timestamps[-1])

    @property
    def track_time(self):
        return np.asarray([
            (stamp - self.timestamps[0]).total_seconds()
            for stamp in self.timestamps
        ])

    def __len__(self):
        return self._timestamps.size

    def copy(self, deep=False):
        obj = super().copy()
        obj._timestamps = self._timestamps
        return obj

    @abc.abstractmethod
    def __getitem__(self, key):
        if isinstance(key, TimeWindow):
            key = slice(key.start, key.stop)

        if isinstance(key, slice):
            start, stop = key.start, key.stop
            if isinstance(start, datetime.datetime):
                start = bisect.bisect_left(self._timestamps, start)
            if isinstance(stop, datetime.datetime):
                stop = bisect.bisect_right(self._timestamps, stop)
            return slice(start, stop)

        if isinstance(key, datetime.datetime):
            idx = bisect.bisect_right(self._timestamps, key)
            if idx == 0:
                return idx
            if idx == len(self):
                return idx - 1
            right_distance = (self._timestamps[idx] - key).total_seconds()
            left_distance = (key - self._timestamps[idx - 1]).total_seconds()
            if left_distance < right_distance:
                idx -= 1
            return idx

        return key


class ReferencedTrack(Track):
    def __init__(self, times, reference):
        self._times = times
        self._reference = reference

    def copy(self, deep=False):
        obj = super().copy(deep=deep)
        obj._times = self._times
        obj._reference = self._reference

    def __len__(self):
        return len(self._times)

    @property
    def track_time(self):
        return self._times - self._times[0]

    @property
    def timestamps(self):
        return [
            self._reference + datetime.timedelta(seconds=t)
            for t in self._times
        ]

    @property
    def time_window(self):
        return TimeWindow(
            start=self._reference + datetime.timedelta(seconds=self._times[0]),
            stop=self._reference + datetime.timedelta(seconds=self._times[-1]),
        )

    @abc.abstractmethod
    def __getitem__(self, key):
        if isinstance(key, TimeWindow):
            key = slice(key.start, key.stop)

        if isinstance(key, slice):
            start, stop = key.start, key.stop
            if isinstance(start, datetime.datetime):
                start = (start - self._reference).total_seconds()
                start = bisect.bisect_left(self._times, start)
            if isinstance(stop, datetime.datetime):
                stop = (stop - self._reference).total_seconds()
                stop = bisect.bisect_right(self._times, stop)
            return slice(start, stop)

        if isinstance(key, datetime.datetime):
            key = (key - self._reference).total_seconds()
            idx = bisect.bisect_right(self._times, key)
            if idx == 0:
                return idx
            if idx == len(self):
                return idx - 1
            right_distance = self._times[idx] - key
            left_distance = key - self._times[idx - 1]
            if left_distance < right_distance:
                idx -= 1
            return idx

        return key


class SampledTrack(Track):
    def __init__(self, sampletime, start, num_samples):
        self._sampletime = sampletime
        self._start = start
        self._num_samples = num_samples

    def copy(self, deep=False):
        obj = super().copy(deep=deep)
        obj._sampletime = self._sampletime
        obj._start = self._start
        obj._num_samples = self._num_samples
        return obj

    def __len__(self):
        return self._num_samples

    @property
    def timestamps(self):
        return [
            self._start + datetime.timedelta(seconds=idx * self._sampletime)
            for idx in range(self._num_samples)
        ]

    @property
    def track_time(self):
        return np.ararnge(self._num_samples) * self._sampletime

    @property
    @abc.abstractmethod
    def time_window(self):
        return TimeWindow(
            start=self._start,
            duration=self._num_samples * self._sampletime
        )

    @abc.abstractmethod
    def __getitem__(self, key):
        if isinstance(key, TimeWindow):
            key = slice(key.start, key.stop)

        if isinstance(key, slice):
            start, stop = key.start, key.stop
            if isinstance(start, datetime.datetime):
                start = (start - self._start).total_seconds()
                start = np.math.ceil(start / self._sampletime)
            if isinstance(stop, datetime.datetime):
                stop = (stop - self._start).total_seconds()
                stop = np.math.floor(stop / self._sampletime)
            return slice(start, stop)

        if isinstance(key, datetime.datetime):
            key = (key - self._start).total_seconds()
            idx = round(key * self._sampletime)
            return idx

        return key


class GPXTrack(TimestampedTrack):
    def __init__(self, path):
        import gpxpy
        file = open(path, 'r')
        contents = gpxpy.parse(file)
        latitudes = []
        longitudes = []
        times = []
        for point in contents.get_points_data():
            latitudes.append(point.point.latitude)
            longitudes.append(point.point.longitude)
            times.append(point.point.time)
        super().__init__(timestamps=times)
        self._latitude = np.asarray(latitudes)
        self._longitude = np.asarray(longitudes)

    @property
    def latitude(self):
        return self._latitude

    @property
    def longitude(self):
        return self._longitude

    @property
    def heading(self):
        raise NotImplementedError()

    @property
    def speed(self):
        raise NotImplementedError()

    def copy(self, deep=False):
        obj = super().copy(deep=deep)
        obj._latitude = self._latitude
        obj._longitude = self._longitude
        return obj

    def __getitem__(self, key):
        key = super().__getitem__(key)
        latitude = self.latitude[key]
        longitude = self.longitude[key]
        timestamps = self.timestamps[key]

        if not isinstance(key, slice):
            return Position(
                latitude=latitude,
                longitude=longitude,
                timestamp=timestamps,
            )

        obj = self.copy(deep=False)
        obj._latitude = latitude
        obj._longitude = longitude
        obj._timestamps = timestamps
        return obj


class Blueflow(TimestampedTrack):
    def __init__(self, path):
        import pandas
        self.data = pandas.read_excel(path)

        super().__init__(timestamps=self.data['Timestamp [UTC]'].dt.tz_localize(dateutil.tz.UTC).dt.to_pydatetime())
        # times = df['Timestamp [UTC]'].dt.to_pydatetime()
        # times = timestamps.TimedstampedSequence(times)
        # latitudes = df['Latitude [deg]'].to_numpy()
        # longitudes = df['Longitude [deg]'].to_numpy()
        # track = Track(latitude=latitudes, longitude=longitudes, time=times)
        # track.blueflow = df

    def copy(self, deep=False):
        obj = super().copy(deep=deep)
        obj.data = self.data
        return obj

    @property
    def latitude(self):
        latitude = self.data['Latitude [deg]']
        try:
            return latitude.to_numpy()
        except AttributeError:
            return latitude

    @property
    def longitude(self):
        longitude = self.data['Longitude [deg]']
        try:
            return longitude.to_numpy()
        except AttributeError:
            return longitude

    @property
    def heading(self):
        heading = self.data['Heading [deg]']
        try:
            return heading.to_numpy()
        except AttributeError:
            return heading

    @property
    def speed(self):
        speed = self.data['Speed over ground [kts]'] / one_knot
        try:
            return speed.to_numpy()
        except AttributeError:
            return speed

    def __getitem__(self, key):
        key = super().__getitem__(key)
        obj = self.copy(deep=False)
        obj.data = self.data.iloc[key]
        obj._timestamps = self.timestamps[key]
        if not isinstance(key, slice):
            return Position(
                latitude=obj.latitude,
                longitude=obj.longitude,
                timestamp=obj.timestamps,
            )
        return obj


class ResampledTrack(Track):
    def __init__(self, sampletime, time, latitude, longitude, start_time=None, order='linear', **kwargs):
        self.sampletime = sampletime
        self.order = order
        self.start_time = start_time
        self._unique_times, self._time_indices, self._duplicate_counts = np.unique(time, return_inverse=True, return_counts=True)
        n_samples = np.math.floor((self._unique_times[-1] - self._unique_times[0]) / self.sampletime) + 1
        self.time = np.arange(n_samples) * self.sampletime + self._unique_times[0]

        self.latitude = self._interpolate(latitude)
        self.longitude = self._interpolate(longitude)
        self._extra_data = list(kwargs.keys())
        for key, data in kwargs.items():
            setattr(self, key, self._interpolate(data))

    def _interpolate(self, data):
        if self._unique_times.size < self._time_indices.size:
            data = np.bincount(self._time_indices, data) / self._duplicate_counts
        interpolator = scipy.interpolate.interp1d(self._unique_times, data, kind=self.order)
        return interpolator(self.time)

    def smooth(self, time_constant, smoothing_method='median'):
        return SmoothTrack(
            time_constant=time_constant,
            sampletime=self.sampletime,
            smoothing_method=smoothing_method,
            start_time=self.start_time,
            latitude=self.latitude,
            longitude=self.longitude,
            **{key: getattr(self, key) for key in self._extra_data}
        )


class SmoothTrack(Track):
    def __init__(self, time_constant, sampletime, latitude, longitude, start_time=None, smoothing_method='median', **kwargs):
        self.time_constant = time_constant
        self.sampletime = sampletime
        self.start_time = start_time
        self.smoothing_method = smoothing_method

        self.latitude = self._smooth(latitude)
        self.longitude = self._smooth(longitude)
        for key, data in kwargs.items():
            setattr(self, key, self._smooth(data))

    def _smooth(self, data):
        if self.smoothing_method.lower() == 'median':
            smoother =  self._median_filter
        elif callable(self.smoothing_method):
            smoother = self.smoothing_method
        else:
            raise ValueError(f'Unknown smoothing method {self.smoothing_method}')
        return smoother(data=data, time_constant=self.time_constant, sampletime=self.sampletime)

    @staticmethod
    def _median_filter(data, time_constant, sampletime):
        import scipy.ndimage
        kernel = round(time_constant / sampletime)
        kernel = (1,) * (np.ndim(data) - 1) + (kernel, )
        return scipy.ndimage.percentile_filter(data, 50, kernel)


# @_callable_property(Track, 'resample')
# class ResampledTrack(Track):
#     def __init__(self, track):
#         self._track = track

#     def __call__(self, sampletime, start=None, stop=None, order='linear'):
#         """Resample a position track.

#         Parameters
#         ----------
#         sampletime : numerical
#             The desired time between samples of the track.
#         start : numerical, optional
#             The time in the track where the resampling should start.
#             Default to use the entire track.
#         stop : numerical, optional
#             The time in the track where the resampling should stop.
#             Default to use the entire track.
#         order : integer or string, default "linear"
#             Sets the polynomial order of the interpolation.
#             See `kind` argument of scipy.interpolate.interp1d`
#         """
#         # Handling uniqueness of samples.
#         unique_times, time_indices, duplicate_counts = np.unique(self._track.time, return_inverse=True, return_counts=True)
#         latitude = self._track.latitude
#         longitude = self._track.longitude
#         if unique_times.size < time_indices.size:
#             # We have duplicate timestamps
#             # bincount will count how often each value in `time_indices` occurs
#             # i.e. how many times the same time exists in the time vector.
#             # These values are weighted by the input values in latitude and longitude,
#             # given as the second argument.
#             # Finally the mean is calculated by dividing by the number of duplicates.
#             latitude = np.bincount(time_indices, latitude) / duplicate_counts
#             longitude = np.bincount(time_indices, longitude) / duplicate_counts

#         # Handling subsampling of the domain
#         if start is not None:
#             # Get the index which is just to the right of the start value, then go left one.
#             # This is to include at one additional sample if the requested time is not actually in the times.
#             start_idx = max(np.searchsorted(unique_times, start, 'right') - 1, 0)
#         else:
#             start_idx = 0
#             start = unique_times[0]

#         if stop is not None:
#             # Get the index which is just to the left of the start value, then go right one.
#             # This is to include at one additional sample if the requested time is not actually in the times.
#             stop_idx = min(np.searchsorted(unique_times, stop, 'left') + 1, len(unique_times))
#         else:
#             stop_idx = len(unique_times)
#             stop = unique_times[-1]
#         unique_times = unique_times[start_idx:stop_idx]
#         latitude = latitude[start_idx:stop_idx]
#         longitude = longitude[start_idx:stop_idx]

#         # Performing the interpolation
#         n_samples = np.math.floor((stop - start) / sampletime) + 1
#         time = np.arange(n_samples) * sampletime + start
#         latitude = scipy.interpolate.interp1d(unique_times, latitude, kind=order, bounds_error=False, fill_value=(latitude[0], latitude[-1]))
#         longitude = scipy.interpolate.interp1d(unique_times, longitude, kind=order, bounds_error=False, fill_value=(longitude[0], longitude[-1]))
#         latitude = latitude(time)
#         longitude = longitude(time)

#         # Store the new data
#         self._latitude = latitude
#         self._longitude = longitude
#         self._time = time
#         self.sampletime = sampletime
#         return self


# @_callable_property(Track, 'smooth')
# class SmoothTrack(Track):
#     # This should probably only be creatable from a resampled track, to ensure that we have
#     # consistent timings. If you want to smooth data which is already sampled nicely,
#     # just "resample" it using the same sampling time?
#     def __init__(self, track):
#         self._track = track

#     @property
#     def sampletime(self):
#         try:
#             return self._track.sampletime
#         except AttributeError:
#             raise TypeError('Cannot smooth a track which has no fixed sampling time!')

#     def __call__(self, time_constant, smoothing_method='median'):
#         """Smooths a position track.

#         Parameters
#         ----------
#         time_constant : numerical
#             Controlls the strength of the smoothing. Is in units of seconds.
#         smoothing_method : string, default "median"
#             Sets the smoothing method
#         """
#         self.time_constant = time_constant
#         self.smoothing_method = smoothing_method

#         latitude = self._track.latitude
#         longitude = self._track.longitude

#         if smoothing_method.lower() == 'median':
#             kernel = round(self.time_constant / self.sampletime)
#             kernel += (kernel + 1) % 2  # Median filter kernel has to be odd!

#             smooth_latitude = scipy.signal.medfilt(latitude, kernel)
#             smooth_longitude = scipy.signal.medfilt(longitude, kernel)

#             # Recalculate the edges with a shrinking kernel instead of zero-padding
#             kernel_offset = kernel // 2 + 1
#             for idx in range(kernel // 2):
#                 smooth_latitude[idx] = np.median(latitude[:idx + kernel_offset])
#                 smooth_longitude[idx] = np.median(longitude[:idx + kernel_offset])
#                 smooth_latitude[-1 - idx] = np.median(latitude[-idx - kernel_offset:])
#                 smooth_longitude[-1 - idx] = np.median(longitude[idx - kernel_offset:])
#         else:
#             raise ValueError(f'Unknown smoothing method {smoothing_method}')

#         self._latitude = smooth_latitude
#         self._longitude = smooth_longitude
#         self._time = self._track.time
#         return self



import itertools
import time
import click
import sounddevice as sd
import numpy as np
import pendulum
from pathlib import Path
import toml

def decidecade_center(f):
    return 10**(np.round(np.log10(f) * 10) / 10)


sweeps = {
    "316Hz decidecade": {
        "lower": decidecade_center(316) * 10**(-1 / 20),
        "upper": decidecade_center(316) * 10**(1 / 20),
        "duration": 4,
        "prerecord": 50e-3,
        "postrecord": 150e-3,
        "fade": 25e-3,
    },
    "1kHz decade": {
        "lower": 1e3 * 10**(-1 / 2),
        "upper": 1e3 * 10**(1 / 2),
        "duration": 10,
        "prerecord": 50e-3,
        "postrecord": 150e-3,
        "fade": 25e-3,
    }
}


class SweepMeasurement:
    _log_file = 'measurerement_log.csv'

    def __init__(self,
        output_folder,
        audio_device,
        samplerate,
        output_channels,
        input_channels,
        repeats,
        repeat_delay,
        cycle,
        sweep_sequence,
        sequence_delay,
        input_gain,
        output_gain,
    ):
        self.output_folder = Path(output_folder)
        if not self.output_folder.exists():
            self.output_folder.mkdir(parents=True)
        self.measurement_log = self.output_folder / "measurement_log.toml"

        self.audio_device = audio_device
        self.samplerate = samplerate
        self.input_channels = input_channels
        self.output_channels = output_channels
        self.input_gain = input_gain
        self.output_gain = output_gain

        self.repeats = repeats
        self.repeat_delay = repeat_delay
        self.sweep_sequence = sweep_sequence
        self.sequence_delay = sequence_delay
        self.cycle = cycle

    def write_to_log(self, category, data):
        with open(self.measurement_log, 'a') as logfile:
            toml.dump(
                {
                    category: [data],
                },
                logfile,
            )

    @property
    def audio_device_index(self):
        for device in sd.query_devices():
            if self.audio_device in device["name"]:
                return device["index"]

    def playrec(self, output_signal):
        # return output_signal
        return sd.playrec(
            data=output_signal,
            samplerate=self.samplerate,
            device=self.audio_device_index,
            input_mapping=self.input_channels,
            output_mapping=self.output_channels,
            blocking=True,
        ).transpose()

    def play(self, output_signal):
        sd.play(
            data=output_signal,
            samplerate=self.samplerate,
            device=self.audio_device_index,
            mapping=self.output_channels,
            blocking=True,
        )

    def create_sweep(self, sweep, repeats=1, repeat_delay=0):
        if sweep in sweeps:
            sweep = sweeps[sweep]
        else:
            raise ValueError(f"Unknown sweep `{sweep}`")

        k = round(sweep["lower"] * sweep["duration"] / np.log(sweep["upper"] / sweep["lower"]))
        sweep_duration = k * np.log(sweep["upper"] / sweep["lower"]) / sweep["lower"]
        sweep_samples = round(sweep_duration * self.samplerate)
        phase_rate = sweep_duration / np.log(sweep["upper"] / sweep["lower"])

        t = np.arange(sweep_samples) / self.samplerate
        sweep_signal = np.sin(2 * np.pi * sweep["lower"] * phase_rate * np.exp(t / phase_rate))

        prerecord_samples = round(sweep["prerecord"] * self.samplerate)
        fade_samples = round(sweep["fade"] * self.samplerate)
        postrecord_samples = round(sweep["postrecord"] * self.samplerate)
        pause_samples = round(repeat_delay * self.samplerate)

        total_samples = prerecord_samples + postrecord_samples + sweep_signal.size * repeats + pause_samples * (repeats - 1)

        sweep_signal[:fade_samples] *= np.sin(np.linspace(0, np.pi / 2, fade_samples))**2
        sweep_signal[-fade_samples:] *= np.sin(np.linspace(np.pi / 2, 0, fade_samples))**2

        repeat_samples = sweep_signal.size + pause_samples
        signal = np.zeros(total_samples)
        for repeat in range(repeats):
            start_idx = prerecord_samples + repeat * repeat_samples
            signal[start_idx:start_idx + sweep_signal.size] = sweep_signal

        return signal

    # def measure_one_sweep(self, sweep):
        # self.write_to_log("measurement.sweep", {"sweep": sweep})

    def run_measurement(self):
        self.write_to_log("measurement", {
            "input_channels": self.input_channels,
            "output_channels": self.output_channels,
            "samplerate": self.samplerate,
            "input_gain": self.input_gain,
            "output_gain": self.output_gain,
            "sequence": self.sweep_sequence,
            "cycle": self.cycle,
            "repeats": self.repeats,
            "repeat_delay": self.repeat_delay,
            "sequence_delay": self.sequence_delay,
        })

        for sweep in (itertools.cycle(self.sweep_sequence) if self.cycle else self.sweep_sequence):
            signal = self.create_sweep(sweep, self.repeats, self.repeat_delay)
            now = pendulum.now(tz="UTC").format("YYYY-MM-DD HH-mm-ss")
            click.echo(f"Staring sweep {sweep} at {now}")
            self.write_to_log("measurement.sweep", {"sweep": sweep, "time": now})
            if len(self.input_channels):
                response = self.playrec(signal)
                np.save(self.output_folder / f"{now}.npy", response)
            else:
                self.play(signal)
            time.sleep(self.sequence_delay)


@click.command()
@click.argument("output-folder")
@click.option("--audio-device", default="ASIO MADIface USB", type=str)
@click.option("--samplerate", default=192000, type=int)
@click.option("--output", "-o", "output_channels", multiple=True, help="Output channels used on the audio interface", type=int)
@click.option("--input", "-i", "input_channels", multiple=True, help="Input channels used on the audio interface", type=int)
@click.option("--output-gain", "output_gain", multiple=True, help="Gain used for the output channels on the audio interface", type=float)
@click.option("--input-gain", "input_gain", multiple=True, help="Gain used for the input channels used on the audio interface", type=float)
@click.option("--repeats", type=int, default=1, help="How many times in a row to send each sweep.")
@click.option("--repeat-delay", type=float, default=0.5, help="The delay (in seconds) between each sweep repetition.")
@click.option("--cycle", flag_value=True, help="Give this flag to cycle through the supplied sweeps.")
@click.option("--sequence-delay", type=float, default=2, help="How long to pause after each sweep group.")
@click.option("--sweep", "-s", "sweep_sequence", multiple=True, help="Name of a sweep to send", type=click.Choice(list(sweeps)))
def send_sweeps_cli(
        output_folder,
        audio_device,
        samplerate,
        output_channels,
        input_channels,
        repeats,
        repeat_delay,
        cycle,
        sweep_sequence,
        sequence_delay,
        input_gain,
        output_gain,
):
    measurement_runner = SweepMeasurement(
        output_folder=output_folder,
        audio_device=audio_device,
        samplerate=samplerate,
        output_channels=output_channels,
        input_channels=input_channels,
        repeats=repeats,
        repeat_delay=repeat_delay,
        cycle=cycle,
        sweep_sequence=sweep_sequence,
        sequence_delay=sequence_delay,
        input_gain=input_gain,
        output_gain=output_gain,
    )
    measurement_runner.run_measurement()

if __name__ == "__main__":
    send_sweeps_cli()